package explicit;

import explicit.rewards.FDCTMCRewardsSimple;
import explicit.rewards.MDPRewardsSimple;
import explicit.rewards.StateRewardsArray;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import prism.PrismComponent;
import prism.PrismException;
import prism.PrismPrintStreamLog;
import strat.MDStrategy;
//import cern.colt.*;
//
//import org.apache.commons.math3.*;
//import org.apache.commons.math3.analysis.solvers.LaguerreSolver;
//import org.apache.commons.math3.complex.Complex;

//import org.apache.commons.math3.analysis.solvers;
//import org.apache.commons.math3.complex.Complex;

public class FDCTMCSynthesis extends PrismComponent 
{

	private FDCTMCSimple model;
	private FDCTMCRewardsSimple fDCTMCRewards; 
	private List<Integer> targetStatesFDCTMC;

	private MDPSimple mdp;
	private MDPRewardsSimple mdpRewards;
	private DTMCSimple dTMC; //for new method
	private StateRewardsArray dTMCRewards; //for new method 
	private BitSet target;
	
	private Map<Integer,Integer> fDCTMCStateToFDIndex;
	private ArrayList<Set<Integer>> fDIndexToPotato; //FD to potato 
	private ArrayList<Set<Integer>> fDIndexToOutStates;
	private int[] fDIndexToEntry; //FD to entry
	private int[] fDCTMCStateToMDPState;
	private int[] mDPStateToFDCTMCState;

	private double[] fDIndexToProbReachOutFromEntryThroughExp;
	
	//for computation of one-step rewards and MDP trans. probs. 
	private double[][] fDIndexToFDTransRew, fDIndexToExpTransRew, fDIndexToRateRew;
	//exp transitions
	int[][][] fDIndexToExpToStates;
	double[][][] fDIndexToExpToProbs;
	//FD transitions
	int[][][] fDIndexToFDToStates;
	double[][][] fDIndexToFDToProbs;
	int[][] fDIndexToDTMCStatesToFDCTMCState;	
	
	double[] uniformizationRates;
        
    public FDCTMCSynthesis() {
        mainLog = new PrismPrintStreamLog(System.out);;
    }
    
    public FDCTMCSynthesis(PrismComponent parent) {
        super(parent);
    }
	
    public void computeReachReward(FDCTMCSimple model, FDCTMCRewardsSimple fDCTMCRewards, 
			List<Integer> targetStatesFDCTMC, double error) throws PrismException{
		this.model = model;
		this.fDCTMCRewards = fDCTMCRewards;
		this.targetStatesFDCTMC = targetStatesFDCTMC;

		if(targetStatesFDCTMC.isEmpty())
			throw new PrismException("Set of reachable target states is empty!");
		
		initialise();

		mainLog.println("num of states: " + mdp.getNumStates());
		
		target = new BitSet(mdp.getNumStates());
		for (int i :targetStatesFDCTMC) {
    		target.set(fDCTMCStateToMDPState[i], true);
		}
		/////////////////////add choices to MDP (fill one-step rewards and transition probabilities)
		
		//first fill values for non-potato states
		processNonPotatoStates();

		// pre-process potato states (create subordinated MCs)
		for(int i=0;i<fDIndexToEntry.length;i++) {
			initialisePotatoState(i, fDIndexToEntry[i]);
		}
		
		MDPModelChecker mDPModelChecker = new MDPModelChecker(this); 
		ModelCheckerResult result = null;
		
		//we use initial values FDs to compute corresponding one-step rewards and transition probabilities
		//we fill corresponding MDP and solve it
		double[] temp = new double[1];
		for(int i=0;i<fDIndexToEntry.length;i++) {
			processPotatoState(i, fDIndexToEntry[i], model.getFDEvent(i).getDelayTime(),
					model.getFDEvent(i).getDelayTime(), 1e-11, temp, true); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
		}
	
		//model check created MDP
		
		mainLog.println("Created DTMC for computation of expected cost:\n" + mdp);
		mainLog.println("Created DTMC costs:\n" + mdpRewards+ "\n");
		mainLog.println("Created DTMC target states:\n" + target+ "\n");
		
		result = mDPModelChecker.computeReachRewards(mdp, mdpRewards, target, true, null, null);
		mainLog.println("EXPECTED REACHABILITY REWARD FROM THE INITIAL STATE: " + result.soln[mdp.getFirstInitialState()]);
    }
    
	public void computeSynthesis(FDCTMCSimple model, FDCTMCRewardsSimple fDCTMCRewards, 
			List<Integer> targetStatesFDCTMC, double error) throws PrismException{
		this.model = model;
		this.fDCTMCRewards = fDCTMCRewards;
		this.targetStatesFDCTMC = targetStatesFDCTMC;

		if(targetStatesFDCTMC.isEmpty())
			throw new PrismException("Set of reachable target states is empty!");
		
		
		initialise();

		mainLog.println("num of states: " + mdp.getNumStates());
		
		target = new BitSet(mdp.getNumStates());
		for (int i :targetStatesFDCTMC) {
    		target.set(fDCTMCStateToMDPState[i], true);
		}
		/////////////////////add choices to MDP (fill one-step rewards and transition probabilities)
		
		//first fill values for non-potato states
		processNonPotatoStates();

		// pre-process potato states (create subordinated MCs)
		for(int i=0;i<fDIndexToEntry.length;i++) {
			initialisePotatoState(i, fDIndexToEntry[i]);
		}

	
		mainLog.println("============ initialisation done ==============");

//		hybridDiscretizationMethod2(error);
//		hybridDiscretizationMethod3(error);
//		polynomMethod(error);
		//		for(int i = 0;i<array.length; i++)
//			System.out.print(array[i] + ", ");
//		System.out.println(".");
		

		

//		System.out.println("============================================== old method");		
//		
		discretizationMethod(error);
		
	}

	public double[] polynomMethod(double error) throws PrismException {
		double lambda = 1.3888;
		double dMax = 28.279;
		int numOfSteps = 90;

		
		
		double[] coefficients = new double[numOfSteps];
		
		double lambdaI = -lambda;
		coefficients[0] = lambdaI * 0.36150624539843823;
		for(int i =1; i < numOfSteps; i++) {
			lambdaI = lambdaI *lambda/i;
			coefficients[i] = lambdaI * 0.08337959999999998;
		}
		
		
		lambdaI = lambda;
		coefficients[0] = lambdaI * 0.08337959999999998;
		for(int i =1; i < numOfSteps; i++) {
			lambdaI = lambdaI *lambda/i;
			coefficients[i] = lambdaI * 0.08337959999999998;
		}
		
		System.out.print(coefficients[0]);
		for(int i=1;i<numOfSteps;i++){
			System.out.print(" + " +coefficients[i] + "x^" + i);
		}
		System.out.println("\n.");
		
//		LaguerreSolver solver;
//		solver = new LaguerreSolver(1e-6);
//		
//		for(int i=numOfSteps-1;i>=0;i--){
//			System.out.print(coefficients[i] + " ");
//		}
//		System.out.println("\n.");
		
//		Complex[] roots = solver.solveAllComplex(coefficients, 1.0);
//		
//		for(int i=0;i<roots.length;i++)
//			System.out.println("root " + i + ": " + roots[0]);
		
		return new double[3];
	}
	

	public double[] hybridDiscretizationMethod3(double error) throws PrismException{
		DTMCModelChecker dTMCModelChecker = new DTMCModelChecker(this);
		
		//////// begin of computation of delta, dMax, and kappa
		
//		MDPModelChecker mDPModelChecker = new MDPModelChecker(); 
		ModelCheckerResult result = null;
		
		//we use initial values FDs to compute corresponding one-step rewards and transition probabilities
		//we fill corresponding MDP and solve it
		Distribution distr;
		double[] temp = new double[1];
		for(int i=0;i<fDIndexToEntry.length;i++) {
			distr = processPotatoState(i, fDIndexToEntry[i], model.getFDEvent(i).getDelayTime(),
					model.getFDEvent(i).getDelayTime(), 1e-11, temp, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error

			dTMC.setTransitions(fDCTMCStateToMDPState[fDIndexToEntry[i]], distr);
			dTMCRewards.setStateReward(fDCTMCStateToMDPState[fDIndexToEntry[i]], temp[0]);
		}
	
		//model check created MDP
		
		mainLog.println("Created DTMC for computation of expected cost:\n" + mdp);
		mainLog.println("Created DTMC costs:\n" + mdpRewards+ "\n");
		mainLog.println("Created DTMC target states:\n" + target+ "\n");
		


	
		result = dTMCModelChecker.computeReachRewards(dTMC, dTMCRewards, target, null, null);
		
//		result = mDPModelChecker.computeReachRewards(mdp, mdpRewards, target, true, null, null);		
		
		//computation of relevant constants
		double boundExpectedCost;
		double boundNumOfSteps;
		double alpha;
		double d1;

				
		//////////// boundExpectedCost
		
		//we evaluate bound on expected cost from every state
		//take maximum from all MDP states
		boundExpectedCost = 0.0;
		for(int i=0; i<mdp.getNumStates();i++){
			if(boundExpectedCost < result.soln[i]) 
				boundExpectedCost = result.soln[i];
		}
		
		mainLog.println("\n=========================================\n");
		mainLog.println("BOUND ON EXPECTED COST FROM EACH STATE: " + boundExpectedCost);
		mainLog.println("EXPECTED COST FROM INITIAL STATE: " + result.soln[mdp.getFirstInitialState()]);

		///////// boundNumOfSteps
		
		//we check whether each entry of potato has FD transition with positive impulse cost 
		double minimalOneStepCost = Double.MAX_VALUE;
		for(int i = 0; i<mDPStateToFDCTMCState.length; i++){
			int state = mDPStateToFDCTMCState[i];
			
			if(targetStatesFDCTMC.contains(state)) continue; //skip target states

			double currentCost = 0.0;
			//check whether there is FD event and process accordingly
			if(!fDCTMCStateToFDIndex.containsKey(state)) { 
				//non FD state
//				//first compute impulse costs
//				Iterator<Entry<Integer,Double>> it2;
//				it2 = model.getTransitionsIterator(state);
//				while(it2.hasNext()) {
//					Entry<Integer,Double> entry2 =it2.next();
//					currentCost = currentCost
//										+ entry2.getValue() * fDCTMCRewards.getTransitionReward(-1,
//												state, entry2.getKey());
//	                        }
//				//add rate costs
//				currentCost = currentCost + fDCTMCRewards.getStateReward(state)/model.getExitRate(state);
//
//				//TODO I believe that the above has to be done somewhere also
//				System.out.println(currentCost +" vs. " + mdpRewards.getTransitionReward(fDCTMCStateToMDPState[state], 0)); 
//				
//				//process current minimal cost 
				if(minimalOneStepCost > mdpRewards.getTransitionReward(fDCTMCStateToMDPState[state], 0))
					minimalOneStepCost = mdpRewards.getTransitionReward(fDCTMCStateToMDPState[state], 0);
		
			} else {
				//there is active FD event
				
				//the result one step cost is minimum of FD impulse cost and
				//exp impulse cost and rate costs accumulated in potato
				
				//it is possible to traverse potato and compute the cost on exp trans accumulated in potato
				//for now much simpler and correct implementation (probably could be done better)

				Iterator<Integer> it3 = fDIndexToPotato.get(fDCTMCStateToFDIndex.get(state)).iterator();
				int state2;
				while(it3.hasNext()) {
					state2 = it3.next();

					//first process the impulse cost of FD event
					Iterator<Entry<Integer,Double>> it;
					it = model.getFDEvent(fDCTMCStateToFDIndex.get(state2)).getTransitionsIterator(state2);
					while(it.hasNext()) {
						Entry<Integer,Double> entry2 =it.next();
						currentCost = currentCost 
											+ entry2.getValue() * fDCTMCRewards.getTransitionReward(fDCTMCStateToFDIndex.get(state2), 
													state2 , entry2.getKey());
					}
					if(currentCost<=0)
						throw new PrismException("State " + state2 
								+ " is entry of potato but does not have positive cost (" + currentCost +") on FD transition!");
					if(minimalOneStepCost > currentCost )
						minimalOneStepCost = currentCost;
					
					currentCost = 0.0;
					//first compute impulse costs
					Iterator<Entry<Integer,Double>> it2;
					it2 = model.getTransitionsIterator(state2);
					while(it2.hasNext()) {
						Entry<Integer,Double> entry2 =it2.next();
						currentCost = currentCost
											+ entry2.getValue() * fDCTMCRewards.getTransitionReward(-1,
													state2, entry2.getKey());
		                        }
					//add rate costs
					currentCost = currentCost + fDCTMCRewards.getStateReward(state2)/uniformizationRates[fDCTMCStateToFDIndex.get(state)];
	
					//process current minimal cost 
					if(minimalOneStepCost > currentCost)
						minimalOneStepCost = currentCost;
				}
			}
		}			

//		mainLog.println("MINIMAL ONE STEP COST IS: " + minimalOneStepCost);
		
		boundNumOfSteps = boundExpectedCost/minimalOneStepCost;
	
		mainLog.println("BOUND ON EXPECTED NUMBER OF STEPS FROM EACH STATE: "+ boundNumOfSteps);
		
		
		/////////////// alpha
		
		alpha = error /(boundNumOfSteps * (1+boundExpectedCost*mdp.getNumStates()));
		double alpha2 = 1/(2*boundNumOfSteps*mdp.getNumStates());
		if(alpha2 < alpha ) alpha = alpha2;
		
		mainLog.println("ALPHA: " + alpha);

		mainLog.println("\n============ common bounds computed ==============\n");
		
		long timer;
		double maxStateReward, maxImpulseReward, minStateReward, minBranchingPst, dMax2,maxRateSum, maxRateSumTemp;
		double[] deltas= new double[model.getNumFDEvents()];
		double[] dMaxes= new double[model.getNumFDEvents()];
		double[] kappas= new double[model.getNumFDEvents()];
		int fDIndex;

		for(fDIndex=0;fDIndex<fDIndexToEntry.length;fDIndex++) {
			mainLog.println("Computation of potato for FD " + fDIndex);
			

			////////////// d1
			
			//computation of minimal and maximal nonzero costs/rewards
			maxStateReward = 0.0; // = 2;
			maxImpulseReward = 0.0; // = 2;
			minStateReward = Double.MAX_VALUE; // = 1;
			minBranchingPst = Double.MAX_VALUE; // = 0.5;
			maxRateSum = 0.0;
			
			//exponential transitions
			for(int i : fDIndexToPotato.get(fDIndex)) {
				Iterator<Entry<Integer,Double>> it;
				it = model.getTransitionsIterator(i);
				maxRateSumTemp = 0;
				while(it.hasNext()) {
					Entry<Integer,Double> entry2 =it.next();
					maxRateSumTemp += entry2.getValue(); 
					if(minBranchingPst > entry2.getValue() && entry2.getValue() > 0)
						minBranchingPst = entry2.getValue();
					if(maxImpulseReward < fDCTMCRewards.getTransitionReward(-1,i, entry2.getKey()))
						maxImpulseReward = fDCTMCRewards.getTransitionReward(-1,i, entry2.getKey());
				}
				if(maxRateSum < maxRateSumTemp) 
					maxRateSum = maxRateSumTemp;
			}

			minBranchingPst = minBranchingPst/maxRateSum;
			
			//FD transitions
			for(int i =0; i<model.getNumStates();i++) {
				Iterator<Entry<Integer,Double>> it;
				it = model.getFDEvent(fDIndex).getTransitionsIterator(i);
				while(it.hasNext()) {
					Entry<Integer,Double> entry2 =it.next();
					if(minBranchingPst > entry2.getValue() && entry2.getValue() > 0)
						minBranchingPst = entry2.getValue();
					if(maxImpulseReward < fDCTMCRewards.getTransitionReward(fDIndex,i, entry2.getKey()))
						maxImpulseReward = fDCTMCRewards.getTransitionReward(fDIndex,i, entry2.getKey());
				}
			}

			
			//state rewards
			for(int i =0; i<model.getNumStates();i++) { //TODO is it sufficient to test only states in potato???
				if(minStateReward > fDCTMCRewards.getStateReward(i)&& fDCTMCRewards.getStateReward(i) > 0)
					minStateReward=fDCTMCRewards.getStateReward(i);
				if(maxStateReward < fDCTMCRewards.getStateReward(i))
					maxStateReward = fDCTMCRewards.getStateReward(i);
			}
			
			
			mainLog.println("\n\tmaxStateReward: " + maxStateReward + "\n\tmaxImpulseReward: " + maxImpulseReward 
					+ "\n\tminStateReward: " + minStateReward + "\n\tminBranchingPst: "+ minBranchingPst);
			
			//end of computation of minimal and maximal nonzero costs/rewards
			
			
			d1 = uniformizationRates[fDIndex]* maxStateReward + 2*uniformizationRates[fDIndex] * maxImpulseReward;
			if (d1< 2* uniformizationRates[fDIndex]) d1 = 2* uniformizationRates[fDIndex];
			
			mainLog.println("D1: " + d1);
			
			//compute delta and dmax for every potato    		
			deltas[fDIndex] = alpha/d1;
			kappas[fDIndex] = (error * deltas[fDIndex] * minStateReward)/(2* mdp.getNumStates()*(1+Math.pow(boundExpectedCost,2)));
			//TODO traverse potato and find the minimum probability that some BSCC is reached in potato 
			dMaxes[fDIndex]  = boundExpectedCost/(Math.pow(minBranchingPst, fDIndexToPotato.get(fDIndex).size())*minStateReward);
			dMax2 = (Math.E * Math.abs(Math.log(alpha/2)))/(uniformizationRates[fDIndex]*minBranchingPst);
//			mainLog.println("DMAX: " + dMax + " \tDMaX2: " + dMax2);
			if(dMax2 > dMaxes[fDIndex]) dMaxes[fDIndex] = dMax2;

			
			   			
//    		delta =  0.009;//0.000009;    		
//    		kappa = 1e-9; //TODO pri 14 zacina pocitat extremne pomaly!!! mozno nestaci presnost double!!!
//    		dMax  = 24.8;
    		mainLog.println("\n\tDELTA: " + deltas[fDIndex] + "\n\tKAPPA: " + kappas[fDIndex] + "\n\tD_MAX: "+ dMaxes[fDIndex] + "\n");

    		
    		System.out.println("Num of Poisson steps: " + CbpAlg.getNumberOfPoissonStepsFromError(dMaxes[fDIndex], 1, kappas[fDIndex], uniformizationRates[fDIndex]));
    		mainLog.println("\n============ potato bounds computed ==============\n");
    		
		} //done computation of this potato
		
		
		
		//////// end of computation of delta, dMax, and kappa
		
		ArrayList<Map<Integer,Double>> fdIndexToImpRewFD = new ArrayList<Map<Integer,Double>> (fDIndexToPotato.size());
		double sum;
		Iterator<Entry<Integer, Double>> it;
		Map<Integer,Double> tempMap;
		Entry<Integer, Double> entry;
		for(int i=0;i<fDIndexToPotato.size();i++) {
			tempMap = new HashMap<Integer,Double>(fDIndexToPotato.get(i).size());
			fdIndexToImpRewFD.add(i, tempMap);
			for(int potatoState : fDIndexToPotato.get(i)){
				it = model.getFDEvent(i).getTransitionsIterator(potatoState);
				sum = 0;
				while(it.hasNext()) {
					entry = it.next();
					sum += entry.getValue()*fDCTMCRewards.getTransitionReward(i,potatoState, entry.getKey());
				}
				tempMap.put(potatoState, sum);
			}
		}


		

	
		double[] fdIndexToMinRemainingRew = new double[fDIndexToPotato.size()];		
		double minImpRew = Double.MAX_VALUE;
		double minRateRew = Double.MAX_VALUE;
		for(int i=0;i<fDIndexToPotato.size();i++) {

			for(int potatoState : fDIndexToPotato.get(i)){
				if(minRateRew > fDCTMCRewards.getStateReward(potatoState))
						minRateRew = fDCTMCRewards.getStateReward(potatoState);
				
				it = model.getTransitionsIterator(potatoState);
				sum = 0;
				while(it.hasNext()) {
					entry = it.next();
					sum += entry.getValue()*fDCTMCRewards.getTransitionReward(-1,potatoState, entry.getKey());
				}
				if(minImpRew > sum) minImpRew = sum;
			}
			fdIndexToMinRemainingRew[i] = (1-fDIndexToProbReachOutFromEntryThroughExp[i])*(minRateRew + uniformizationRates[i] * minImpRew);
			System.out.println("FD: " + i  + " minRateRew: "+ minRateRew + " minImpRew: " + minImpRew);

		}
		
		//TODO test whether change of Vector to ArrayList would help, I think the native paralelization thanks to serializable Vector is better
		ArrayList<Vector<Distribution>> fDPositionToDistrs = new ArrayList<Vector<Distribution>>(model.getNumFDEvents());
		ArrayList<Vector<Double>> fDPositionToRewards      = new ArrayList<Vector<Double>>(model.getNumFDEvents()); 
		ArrayList<Vector<Double>> fDPositionToDelays       = new ArrayList<Vector<Double>>(model.getNumFDEvents());
		//TODO the above, when performance will be good can be changed to dedicated object ... holding delay, reward and distribution
		//TODO or to arrays
		ArrayList<LinkedList<Integer>> fDOrderToPosition = new ArrayList<LinkedList<Integer>>(model.getNumFDEvents());
		int[] currentFDPositions = new int[model.getNumFDEvents()];
		
		
		double[] nextReward = new double[1];
		Distribution nextDistr = null;
		//feeding structures with initial vector of FDs minimal and maximal delays
		for(int i=0;i<model.getNumFDEvents();i++){
			fDPositionToDistrs.add(new Vector<Distribution>());
			fDPositionToRewards.add(new Vector<Double>());
			fDPositionToDelays.add(new Vector<Double>());
			fDOrderToPosition.add(new LinkedList<Integer>());

			nextDistr = processPotatoState(i, fDIndexToEntry[i], deltas[i],
					deltas[i], kappas[i], nextReward, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
			
			fDPositionToDistrs.get(i).add(nextDistr);
			fDPositionToRewards.get(i).add(nextReward[0]);
			fDPositionToDelays.get(i).add(deltas[i]);
			fDOrderToPosition.get(i).add(0);
			
			nextDistr = processPotatoState(i, fDIndexToEntry[i], model.getFDEvent(i).getDelayTime(),
					model.getFDEvent(i).getDelayTime(), kappas[i], nextReward, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
			
			fDPositionToDistrs.get(i).add(nextDistr);
			fDPositionToRewards.get(i).add(nextReward[0]);
			fDPositionToDelays.get(i).add(model.getFDEvent(i).getDelayTime());
			fDOrderToPosition.get(i).add(1);

			nextDistr = processPotatoState(i, fDIndexToEntry[i], dMaxes[i],
					dMaxes[i], kappas[i], nextReward, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
			
			fDPositionToDistrs.get(i).add(nextDistr);
			fDPositionToRewards.get(i).add(nextReward[0]);
			fDPositionToDelays.get(i).add(dMaxes[i]);
			fDOrderToPosition.get(i).add(2);
			
			currentFDPositions[i] = 1;
		}


		//TODO in fDIndexToDistrs we have defective distributions (not summing to one) does it cause some error in computation of values?
		
		int mDPState, position, tempPosition, minPosition;
		double[] values = null;
		double minValue, newValue, step, delay;
		boolean doNextStep = true;
		int i;
		Iterator<Entry<Integer,Double>> distrIt;
		Entry<Integer,Double> distrEntry;
		
		
		ListIterator<Integer> listIterator;
		
		int k, j= 0;
		double tempValueMax, tempValueMin, delta2;
		
		while (doNextStep) {
			j++;
			doNextStep = false;
			
			//compute values vector for current delays, i.e. currentDelayIndexes 
			for(i=0;i<model.getNumFDEvents();i++){
				mDPState = fDCTMCStateToMDPState[fDIndexToEntry[i]];
				position = currentFDPositions[i];

				dTMC.setTransitions(mDPState, fDPositionToDistrs.get(i).get(position));
				dTMCRewards.setStateReward(mDPState, fDPositionToRewards.get(i).get(position));
			}
			
			//TODO the setting of values of init and known may cause some speedup, check it up!		
			values = dTMCModelChecker.computeReachRewards(dTMC, dTMCRewards, target, null, null).soln;
			
			
			for(i=0;i<model.getNumFDEvents();i++){
				mDPState = fDCTMCStateToMDPState[fDIndexToEntry[i]];
//				System.out.println("entry:    " + fDIndexToEntry[i] + " val: " + values[mDPState]);
				minPosition = currentFDPositions[i];
				minValue = values[mDPState];

				//compute min and max of values
				tempValueMin = Double.MAX_VALUE; //TODO check whether correct
				tempValueMax = 0.0;
				for(int potatoState : fDIndexToPotato.get(i)){
					it = model.getFDEvent(i).getTransitionsIterator(potatoState);
					sum = 0;
					while(it.hasNext()) {
						entry = it.next();
						sum += entry.getValue()*values[fDCTMCStateToMDPState[entry.getKey()]];
					}
					if(tempValueMax < (sum + fdIndexToImpRewFD.get(i).get(potatoState)))
						tempValueMax = sum + fdIndexToImpRewFD.get(i).get(potatoState);
					if(tempValueMin > (sum + fdIndexToImpRewFD.get(i).get(potatoState)))
						tempValueMax = sum + fdIndexToImpRewFD.get(i).get(potatoState);
				}
				
				for(int state : fDIndexToOutStates.get(i)){
//					System.out.println("outState: " + state + " val: " + values[fDCTMCStateToMDPState[state]]);
					if(tempValueMax < values[fDCTMCStateToMDPState[state]])
						tempValueMax = values[fDCTMCStateToMDPState[state]];
					if(tempValueMin > values[fDCTMCStateToMDPState[state]])
						tempValueMin = values[fDCTMCStateToMDPState[state]];
				}
				delta2 = uniformizationRates[i]*(tempValueMax - tempValueMin) - fdIndexToMinRemainingRew[i];
				if(delta2 <= 0.0)
					delta2 = 0.000001;
				
				listIterator = fDOrderToPosition.get(i).listIterator(0);
				position = listIterator.next();
				k=0;
				while(true){
					k++;
					//compute value for this position
					newValue = fDPositionToRewards.get(i).get(position);
					//multiplication of distribution nextDistr with vector of values
					distrIt = fDPositionToDistrs.get(i).get(position).iterator();
					while(distrIt.hasNext()){
						distrEntry = distrIt.next();
						newValue += distrEntry.getValue() * values[distrEntry.getKey()];
					}
					
//					System.out.println(i +": NEW: " + newValue + " MIN: " + minValue);
					
					if(newValue < minValue){
						doNextStep = true;
						minValue = newValue;
						minPosition = position;
					}
					
					if(!listIterator.hasNext()) break; //this was the last position in list => we reached dMax
					
					step =  deltas[i] + ((newValue-minValue)/delta2);

//					System.out.print(", " + step);
					
					delay = fDPositionToDelays.get(i).get(position) + step; 
					//find the largest position of pre-computed delay that we jump over
					position = -1;
					while(listIterator.hasNext()){
						tempPosition = listIterator.next();
						if(fDPositionToDelays.get(i).get(tempPosition)>delay) {
							listIterator.previous();
							break;
						}
						position = tempPosition;
					}
					
					//there was no delay that we jumped over (iterator is pointing to the current position and the position is not last)
					if(-1 == position) {
						nextDistr = processPotatoState(i, fDIndexToEntry[i], delay,
								delay, kappas[i], nextReward,false);
						//store computed values
						fDPositionToRewards.get(i).add(nextReward[0]);
						fDPositionToDistrs.get(i).add(nextDistr);
						fDPositionToDelays.get(i).add(delay);
						position = fDPositionToDistrs.get(i).size()-1;
						listIterator.add(position);
					}
				}
				System.out.println("3>>>>>>>>> Iteration: " + j + " FD: " + i + " num of steps: " + k + " computed: "  
						+ (fDPositionToDistrs.get(i).size()-1) +" old:" 
						+ (int)Math.floor(dMaxes[i]/deltas[i]) + " FD with min. value: " + fDPositionToDelays.get(i).get(minPosition));
				
				// update current positions
				currentFDPositions[i] = minPosition;
			}
		}

//		System.out.println("DTMC: " + dTMC);
//		for(i = 0;i<values.length; i++)
//			System.out.print(values[i] + ", ");
//		System.out.println(".");
		
		mainLog.print("\nRESULT in initial state: " + values[mdp.getFirstInitialState()] + " delays: ");
		for(i=0;i<fDIndexToEntry.length;i++){
			mainLog.print(model.getFDEvent(i).getLabel() + "=" 
					+ fDPositionToDelays.get(i).get(currentFDPositions[i]) + " ");
		}
		mainLog.println("");

		return values;
	}
	
	public double[] hybridDiscretizationMethod2(double error) throws PrismException{
		DTMCModelChecker dTMCModelChecker = new DTMCModelChecker(this);
		
		//////// begin of computation of delta, dMax, and kappa
		
//		MDPModelChecker mDPModelChecker = new MDPModelChecker(); 
		ModelCheckerResult result = null;
		
		//we use initial values FDs to compute corresponding one-step rewards and transition probabilities
		//we fill corresponding MDP and solve it
		Distribution distr;
		double[] temp = new double[1];
		for(int i=0;i<fDIndexToEntry.length;i++) {
			distr = processPotatoState(i, fDIndexToEntry[i], model.getFDEvent(i).getDelayTime(),
					model.getFDEvent(i).getDelayTime(), 1e-11, temp, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error

			dTMC.setTransitions(fDCTMCStateToMDPState[fDIndexToEntry[i]], distr);
			dTMCRewards.setStateReward(fDCTMCStateToMDPState[fDIndexToEntry[i]], temp[0]);
		}
	
		//model check created MDP
		
		mainLog.println("Created DTMC for computation of expected cost:\n" + mdp);
		mainLog.println("Created DTMC costs:\n" + mdpRewards+ "\n");
		mainLog.println("Created DTMC target states:\n" + target+ "\n");
		


	
		result = dTMCModelChecker.computeReachRewards(dTMC, dTMCRewards, target, null, null);
		
//		result = mDPModelChecker.computeReachRewards(mdp, mdpRewards, target, true, null, null);		
		
		//computation of relevant constants
		double boundExpectedCost;
		double boundNumOfSteps;
		double alpha;
		double d1;

				
		//////////// boundExpectedCost
		
		//we evaluate bound on expected cost from every state
		//take maximum from all MDP states
		boundExpectedCost = 0.0;
		for(int i=0; i<mdp.getNumStates();i++){
			if(boundExpectedCost < result.soln[i]) 
				boundExpectedCost = result.soln[i];
		}
		
		mainLog.println("\n=========================================\n");
		mainLog.println("BOUND ON EXPECTED COST FROM EACH STATE: " + boundExpectedCost);
		mainLog.println("EXPECTED COST FROM INITIAL STATE: " + result.soln[mdp.getFirstInitialState()]);

		///////// boundNumOfSteps
		
		//we check whether each entry of potato has FD transition with positive impulse cost 
		double minimalOneStepCost = Double.MAX_VALUE;
		for(int i = 0; i<mDPStateToFDCTMCState.length; i++){
			int state = mDPStateToFDCTMCState[i];
			
			if(targetStatesFDCTMC.contains(state)) continue; //skip target states

			double currentCost = 0.0;
			//check whether there is FD event and process accordingly
			if(!fDCTMCStateToFDIndex.containsKey(state)) { 
				//non FD state
//				//first compute impulse costs
//				Iterator<Entry<Integer,Double>> it2;
//				it2 = model.getTransitionsIterator(state);
//				while(it2.hasNext()) {
//					Entry<Integer,Double> entry2 =it2.next();
//					currentCost = currentCost
//										+ entry2.getValue() * fDCTMCRewards.getTransitionReward(-1,
//												state, entry2.getKey());
//	                        }
//				//add rate costs
//				currentCost = currentCost + fDCTMCRewards.getStateReward(state)/model.getExitRate(state);
//
//				//TODO I believe that the above has to be done somewhere also
//				System.out.println(currentCost +" vs. " + mdpRewards.getTransitionReward(fDCTMCStateToMDPState[state], 0)); 
//				
//				//process current minimal cost 
				if(minimalOneStepCost > mdpRewards.getTransitionReward(fDCTMCStateToMDPState[state], 0))
					minimalOneStepCost = mdpRewards.getTransitionReward(fDCTMCStateToMDPState[state], 0);
		
			} else {
				//there is active FD event
				
				//the result one step cost is minimum of FD impulse cost and
				//exp impulse cost and rate costs accumulated in potato
				
				//it is possible to traverse potato and compute the cost on exp trans accumulated in potato
				//for now much simpler and correct implementation (probably could be done better)

				Iterator<Integer> it3 = fDIndexToPotato.get(fDCTMCStateToFDIndex.get(state)).iterator();
				int state2;
				while(it3.hasNext()) {
					state2 = it3.next();

					//first process the impulse cost of FD event
					Iterator<Entry<Integer,Double>> it;
					it = model.getFDEvent(fDCTMCStateToFDIndex.get(state2)).getTransitionsIterator(state2);
					while(it.hasNext()) {
						Entry<Integer,Double> entry2 =it.next();
						currentCost = currentCost 
											+ entry2.getValue() * fDCTMCRewards.getTransitionReward(fDCTMCStateToFDIndex.get(state2), 
													state2 , entry2.getKey());
					}
					if(currentCost<=0)
						throw new PrismException("State " + state2 
								+ " is entry of potato but does not have positive cost (" + currentCost +") on FD transition!");
					if(minimalOneStepCost > currentCost )
						minimalOneStepCost = currentCost;
					
					currentCost = 0.0;
					//first compute impulse costs
					Iterator<Entry<Integer,Double>> it2;
					it2 = model.getTransitionsIterator(state2);
					while(it2.hasNext()) {
						Entry<Integer,Double> entry2 =it2.next();
						currentCost = currentCost
											+ entry2.getValue() * fDCTMCRewards.getTransitionReward(-1,
													state2, entry2.getKey());
		                        }
					//add rate costs
					currentCost = currentCost + fDCTMCRewards.getStateReward(state2)/uniformizationRates[fDCTMCStateToFDIndex.get(state)];
	
					//process current minimal cost 
					if(minimalOneStepCost > currentCost)
						minimalOneStepCost = currentCost;
				}
			}
		}			

//		mainLog.println("MINIMAL ONE STEP COST IS: " + minimalOneStepCost);
		
		boundNumOfSteps = boundExpectedCost/minimalOneStepCost;
	
		mainLog.println("BOUND ON EXPECTED NUMBER OF STEPS FROM EACH STATE: "+ boundNumOfSteps);
		
		
		/////////////// alpha
		
		alpha = error /(boundNumOfSteps * (1+boundExpectedCost*mdp.getNumStates()));
		double alpha2 = 1/(2*boundNumOfSteps*mdp.getNumStates());
		if(alpha2 < alpha ) alpha = alpha2;
		
		mainLog.println("ALPHA: " + alpha);

		mainLog.println("\n============ common bounds computed ==============\n");
		
		long timer;
		double maxStateReward, maxImpulseReward, minStateReward, minBranchingPst, dMax2,maxRateSum, maxRateSumTemp;
		double[] deltas= new double[model.getNumFDEvents()];
		double[] dMaxes= new double[model.getNumFDEvents()];
		double[] kappas= new double[model.getNumFDEvents()];
		int fDIndex;

		for(fDIndex=0;fDIndex<fDIndexToEntry.length;fDIndex++) {
			mainLog.println("Computation of potato for FD " + fDIndex);
			

			////////////// d1
			
			//computation of minimal and maximal nonzero costs/rewards
			maxStateReward = 0.0; // = 2;
			maxImpulseReward = 0.0; // = 2;
			minStateReward = Double.MAX_VALUE; // = 1;
			minBranchingPst = Double.MAX_VALUE; // = 0.5;
			maxRateSum = 0.0;
			
			//exponential transitions
			for(int i : fDIndexToPotato.get(fDIndex)) {
				Iterator<Entry<Integer,Double>> it;
				it = model.getTransitionsIterator(i);
				maxRateSumTemp = 0;
				while(it.hasNext()) {
					Entry<Integer,Double> entry2 =it.next();
					maxRateSumTemp += entry2.getValue(); 
					if(minBranchingPst > entry2.getValue() && entry2.getValue() > 0)
						minBranchingPst = entry2.getValue();
					if(maxImpulseReward < fDCTMCRewards.getTransitionReward(-1,i, entry2.getKey()))
						maxImpulseReward = fDCTMCRewards.getTransitionReward(-1,i, entry2.getKey());
				}
				if(maxRateSum < maxRateSumTemp) 
					maxRateSum = maxRateSumTemp;
			}

			minBranchingPst = minBranchingPst/maxRateSum;
			
			//FD transitions
			for(int i =0; i<model.getNumStates();i++) {
				Iterator<Entry<Integer,Double>> it;
				it = model.getFDEvent(fDIndex).getTransitionsIterator(i);
				while(it.hasNext()) {
					Entry<Integer,Double> entry2 =it.next();
					if(minBranchingPst > entry2.getValue() && entry2.getValue() > 0)
						minBranchingPst = entry2.getValue();
					if(maxImpulseReward < fDCTMCRewards.getTransitionReward(fDIndex,i, entry2.getKey()))
						maxImpulseReward = fDCTMCRewards.getTransitionReward(fDIndex,i, entry2.getKey());
				}
			}

			
			//state rewards
			for(int i =0; i<model.getNumStates();i++) { //TODO is it sufficient to test only states in potato???
				if(minStateReward > fDCTMCRewards.getStateReward(i)&& fDCTMCRewards.getStateReward(i) > 0)
					minStateReward=fDCTMCRewards.getStateReward(i);
				if(maxStateReward < fDCTMCRewards.getStateReward(i))
					maxStateReward = fDCTMCRewards.getStateReward(i);
			}
			
			
			mainLog.println("\n\tmaxStateReward: " + maxStateReward + "\n\tmaxImpulseReward: " + maxImpulseReward 
					+ "\n\tminStateReward: " + minStateReward + "\n\tminBranchingPst: "+ minBranchingPst);
			
			//end of computation of minimal and maximal nonzero costs/rewards
			
			
			d1 = uniformizationRates[fDIndex]* maxStateReward + 2*uniformizationRates[fDIndex] * maxImpulseReward;
			if (d1< 2* uniformizationRates[fDIndex]) d1 = 2* uniformizationRates[fDIndex];
			
			mainLog.println("D1: " + d1);
			
			//compute delta and dmax for every potato    		
			deltas[fDIndex] = alpha/d1;
			kappas[fDIndex] = (error * deltas[fDIndex] * minStateReward)/(2* mdp.getNumStates()*(1+Math.pow(boundExpectedCost,2)));
			//TODO traverse potato and find the minimum probability that some BSCC is reached in potato 
			dMaxes[fDIndex]  = boundExpectedCost/(Math.pow(minBranchingPst, fDIndexToPotato.get(fDIndex).size())*minStateReward);
			dMax2 = (Math.E * Math.abs(Math.log(alpha/2)))/(uniformizationRates[fDIndex]*minBranchingPst);
//			mainLog.println("DMAX: " + dMax + " \tDMaX2: " + dMax2);
			if(dMax2 > dMaxes[fDIndex]) dMaxes[fDIndex] = dMax2;

			
			   			
//    		delta =  0.009;//0.000009;    		
//    		kappa = 1e-9; //TODO pri 14 zacina pocitat extremne pomaly!!! mozno nestaci presnost double!!!
//    		dMax  = 24.8;
    		mainLog.println("\n\tDELTA: " + deltas[fDIndex] + "\n\tKAPPA: " + kappas[fDIndex] + "\n\tD_MAX: "+ dMaxes[fDIndex] + "\n");

    		
    		System.out.println("Num of Poisson steps: " + CbpAlg.getNumberOfPoissonStepsFromError(dMaxes[fDIndex], 1, kappas[fDIndex], uniformizationRates[fDIndex]));
    		mainLog.println("\n============ potato bounds computed ==============\n");
    		
		} //done computation of this potato
		
		
		
		//////// end of computation of delta, dMax, and kappa
		
		
		//TODO test whether change of Vector to ArrayList would help, I think the native paralelization thanks to serializable Vector is better
		ArrayList<Vector<Distribution>> fDPositionToDistrs = new ArrayList<Vector<Distribution>>(model.getNumFDEvents());
		ArrayList<Vector<Double>> fDPositionToRewards      = new ArrayList<Vector<Double>>(model.getNumFDEvents()); 
		ArrayList<Vector<Double>> fDPositionToDelays       = new ArrayList<Vector<Double>>(model.getNumFDEvents());
		//TODO the above, when performance will be good can be changed to dedicated object ... holding delay, reward and distribution
		//TODO or to arrays
		ArrayList<LinkedList<Integer>> fDOrderToPosition = new ArrayList<LinkedList<Integer>>(model.getNumFDEvents());
		int[] currentFDPositions = new int[model.getNumFDEvents()];
		
		
		double[] nextReward = new double[1];
		Distribution nextDistr = null;
		//feeding structures with initial vector of FDs minimal and maximal delays
		for(int i=0;i<model.getNumFDEvents();i++){
			fDPositionToDistrs.add(new Vector<Distribution>());
			fDPositionToRewards.add(new Vector<Double>());
			fDPositionToDelays.add(new Vector<Double>());
			fDOrderToPosition.add(new LinkedList<Integer>());

			nextDistr = processPotatoState(i, fDIndexToEntry[i], deltas[i],
					deltas[i], kappas[i], nextReward, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
			
			fDPositionToDistrs.get(i).add(nextDistr);
			fDPositionToRewards.get(i).add(nextReward[0]);
			fDPositionToDelays.get(i).add(deltas[i]);
			fDOrderToPosition.get(i).add(0);
			
			nextDistr = processPotatoState(i, fDIndexToEntry[i], model.getFDEvent(i).getDelayTime(),
					model.getFDEvent(i).getDelayTime(), kappas[i], nextReward, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
			
			fDPositionToDistrs.get(i).add(nextDistr);
			fDPositionToRewards.get(i).add(nextReward[0]);
			fDPositionToDelays.get(i).add(model.getFDEvent(i).getDelayTime());
			fDOrderToPosition.get(i).add(1);

			nextDistr = processPotatoState(i, fDIndexToEntry[i], dMaxes[i],
					dMaxes[i], kappas[i], nextReward, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
			
			fDPositionToDistrs.get(i).add(nextDistr);
			fDPositionToRewards.get(i).add(nextReward[0]);
			fDPositionToDelays.get(i).add(dMaxes[i]);
			fDOrderToPosition.get(i).add(2);
			
			currentFDPositions[i] = 1;
		}


		//TODO in fDIndexToDistrs we have defective distributions (not summing to one) does it cause some error in computation of values?
		
		int mDPState, position, tempPosition, minPosition;
		double[] values = null;
		double minValue, newValue, step, delay;
		boolean doNextStep = true;
		int i;
		Iterator<Entry<Integer,Double>> distrIt;
		Entry<Integer,Double> distrEntry;
		
		
		ListIterator<Integer> listIterator;
		
		int k, j= 0;

		
		while (doNextStep) {
			j++;
			doNextStep = false;
			
			//compute values vector for current delays, i.e. currentDelayIndexes 
			for(i=0;i<model.getNumFDEvents();i++){
				mDPState = fDCTMCStateToMDPState[fDIndexToEntry[i]];
				position = currentFDPositions[i];

				dTMC.setTransitions(mDPState, fDPositionToDistrs.get(i).get(position));
				dTMCRewards.setStateReward(mDPState, fDPositionToRewards.get(i).get(position));
			}
			
			//TODO the setting of values of init and known may cause some speedup, check it up!		
			values = dTMCModelChecker.computeReachRewards(dTMC, dTMCRewards, target, null, null).soln;
			
			for(i=0;i<model.getNumFDEvents();i++){
				mDPState = fDCTMCStateToMDPState[fDIndexToEntry[i]];
				minPosition = currentFDPositions[i];
				minValue = values[mDPState];
				
				listIterator = fDOrderToPosition.get(i).listIterator(0);
				position = listIterator.next();
				k=0;
				while(true){
					k++;
					//compute value for this position
					newValue = fDPositionToRewards.get(i).get(position);
					//multiplication of distribution nextDistr with vector of values
					distrIt = fDPositionToDistrs.get(i).get(position).iterator();
					while(distrIt.hasNext()){
						distrEntry = distrIt.next();
						newValue += distrEntry.getValue() * values[distrEntry.getKey()];
					}
					
//					System.out.println(i +": NEW: " + newValue + " MIN: " + minValue);
					
					if(newValue < minValue){
						doNextStep = true;
						minValue = newValue;
						minPosition = position;
					}
					
					if(!listIterator.hasNext()) break; //this was the last position in list => we reached dMax
					
					step = Math.max((int)Math.ceil((newValue-minValue)/error),1);

//					System.out.print(", " + step);
					
					delay = fDPositionToDelays.get(i).get(position) + step * deltas[i]; 
					//find the largest position of pre-computed delay that we jump over
					position = -1;
					while(listIterator.hasNext()){
						tempPosition = listIterator.next();
						if(fDPositionToDelays.get(i).get(tempPosition)>delay) {
							listIterator.previous();
							break;
						}
						position = tempPosition;
					}
					
					//there was no delay that we jumped over (iterator is pointing to the current position and the position is not last)
					if(-1 == position) {
						nextDistr = processPotatoState(i, fDIndexToEntry[i], delay,
								delay, kappas[i], nextReward,false);
						//store computed values
						fDPositionToRewards.get(i).add(nextReward[0]);
						fDPositionToDistrs.get(i).add(nextDistr);
						fDPositionToDelays.get(i).add(delay);
						position = fDPositionToDistrs.get(i).size()-1;
						listIterator.add(position);
					}
				}
				System.out.println(">>>>>>>>> Iteration: " + j + " FD: " + i + " num of steps: " + k 
						+ " computed: "  + (fDPositionToDistrs.get(i).size()-1) +" old:" + (int)Math.floor(dMaxes[i]/deltas[i])
						+ " FD with min. value: " + fDPositionToDelays.get(i).get(minPosition));
				
				// update current positions
				currentFDPositions[i] = minPosition;
			}
		}

//		System.out.println("DTMC: " + dTMC);
//		for(i = 0;i<values.length; i++)
//			System.out.print(values[i] + ", ");
//		System.out.println(".");
		
		mainLog.print("\nRESULT in initial state: " + values[mdp.getFirstInitialState()] + " delays: ");
		for(i=0;i<fDIndexToEntry.length;i++){
			mainLog.print(model.getFDEvent(i).getLabel() + "=" 
					+ fDPositionToDelays.get(i).get(currentFDPositions[i]) + " ");
		}
		mainLog.println("");

		return values;
	}
	
	//TODO this method so far doesn't use the discretization of each potato separately
	public double[] hybridDiscretizationMethod(double error) throws PrismException{
		double uniformizationRate = model.getMaxExitRate();
		DTMCSimple uniformisedDTMC = model.buildUniformisedDTMC(uniformizationRate);
		
		//TODO test whether cahge of Vector to ArrayList would help, I think the native paralelization thanks to serializable Vector is better
		ArrayList<Vector<Distribution>> fDPositionToDistrs = new ArrayList<Vector<Distribution>>(model.getNumFDEvents());
		ArrayList<Vector<Double>> fDpositionToRewards      = new ArrayList<Vector<Double>>(model.getNumFDEvents()); 
		ArrayList<Vector<Double>> fDPositionToDelays       = new ArrayList<Vector<Double>>(model.getNumFDEvents());
		//TODO the above, when performance will be good can be changed to dedicated object ... holding delay, reward and distribution
		ArrayList<LinkedList<Integer>> fDOrderToPosition = new ArrayList<LinkedList<Integer>>(model.getNumFDEvents());
		int[] currentFDPositions = new int[model.getNumFDEvents()];
		
		DTMCModelChecker dTMCModelChecker = new DTMCModelChecker(this); 
		
		double[] nextReward = new double[1];
		Distribution nextDistr = null;
		//feeding structures with initial vector of FDs
		for(int i=0;i<model.getNumFDEvents();i++){
			fDPositionToDistrs.add(new Vector<Distribution>());
			fDpositionToRewards.add(new Vector<Double>());
			fDPositionToDelays.add(new Vector<Double>());
			fDOrderToPosition.add(new LinkedList<Integer>());
			
			nextDistr = processPotatoState(i, fDIndexToEntry[i], model.getFDEvent(i).getDelayTime(),
					model.getFDEvent(i).getDelayTime(), 1e-11, nextReward, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
			
			fDPositionToDistrs.get(i).add(nextDistr);
			fDpositionToRewards.get(i).add(nextReward[0]);
			fDPositionToDelays.get(i).add(model.getFDEvent(i).getDelayTime());
			fDOrderToPosition.get(i).add(0);

			currentFDPositions[i] = 0;
		}


		//TODO in fDIndexToDistrs we have defective distributions (not summing to one) does it cause some error in computation of values?
		
		//compute values vector for current delays, i.e. currentDelayIndexes 
		
		int mDPState, index;
		for(int i=0;i<model.getNumFDEvents();i++){
			mDPState = fDCTMCStateToMDPState[fDIndexToEntry[i]];
			index = currentFDPositions[i];

			dTMC.setTransitions(mDPState, fDPositionToDistrs.get(i).get(index));
			dTMCRewards.setStateReward(mDPState, fDpositionToRewards.get(i).get(index));
		}
		
		System.out.println("DTMC: " + dTMC);
		
		double[] values= null;
		//TODO the setting of values of init and known may cause some speedup, check it up!		
		values = dTMCModelChecker.computeReachRewards(dTMC, dTMCRewards, target, null, null).soln;

		for(int j = 0;j<values.length; j++)
			System.out.print(values[j] + ", ");
		System.out.println(".");
		
		BitSet remain = new BitSet(dTMC.getNumStates());
		for (int i=0;i<dTMC.getNumStates();i++) {
			remain.set(i, true); //fDCTMCStateToMDPState.get(
		}

		BitSet tempTarget = new BitSet(dTMC.getNumStates());
		for (int i=0;i<dTMC.getNumStates();i++) {
			tempTarget.set(i, false);
		}

		BitSet tempTarget2 = new BitSet(model.getNumStates());
		for (int i=0;i<model.getNumStates();i++) {
			tempTarget2.set(i, true);
		}
		
		int i,j;

		double[] reachProbs2, reachProbs = null;
		double[] fDIndexToMaxFDTransRew  = new double[model.getNumFDEvents()];
		double[] fDIndexToMaxExpTransRew = new double[model.getNumFDEvents()];
		double[] fDIndexToMaxRateRew     = new double[model.getNumFDEvents()];
		double maxValue, reachEntry, reachOutStates, minStateRew, dMax, minDeltaProbOrRew; 
		double delta, currDelay, currReward, nextDelay, currMaxDiff;
		boolean useDMax;
		
		dMax = 0.0; 
		minDeltaProbOrRew = 0.0; 
		currMaxDiff = 1.0;
		
		Distribution currDistr;
		Iterator<Entry<Integer,Double>> distrIt;
		Entry<Integer,Double> distrEntry;		
		
		for(i =0; i<model.getNumFDEvents();i++){
			mDPState = fDCTMCStateToMDPState[fDIndexToEntry[i]];
			
			//compute p(s)(t)
			//use method for computation of reachability probabilities in DTMC where there is self-loop with prob. 1 on s
			// the method should return vector of reachability probabilities of target states from any state
			remain.set(mDPState, false);
			//TODO the setting of values of init and known may cause some speedup, check it up!
			reachProbs = dTMCModelChecker.computeReachProbs(dTMC, remain, target, null, null).soln;
			remain.set(mDPState, true);

			//compute probability of reaching mDPState from initial state, i.e. reachEntry
//			tempTarget.set(mDPState, true);
//			//TODO the setting of values of init and known may cause some speedup, check it up!
////			result = dTMCModelChecker.computeReachProbs(dTMC, null, tempTarget, null, null); //TODO maybe it will be faster not to compute exact reachEntry and over-estimate it with 1.0. TEST IT!
//			reachEntry = dTMCModelChecker.computeReachProbs(dTMC, null, tempTarget, null, null).soln[0];
//			tempTarget.set(mDPState, false); //TODO erase
			reachEntry = 1-reachProbs[0];
			
			System.out.println("======= state: " + mDPState);
			for(j = 0;j<reachProbs.length; j++)
				System.out.print(reachProbs[j] + ", ");
			System.out.println(".");

//			for(int j = 0;j<result.soln.length; j++)
//				System.out.print(result.soln[j] + ", ");
//			System.out.println(".");

			
			//discretization 
			//first compute necessary constants
			//compute fDIndexToMaxFDTransRew, maxRateRew and maxExpTransRew
			maxValue = 0.0;
			for(j = 0;j<values.length;j++){
				if(maxValue< values[j])
					maxValue= values[j];
			}
			
			fDIndexToMaxFDTransRew[i] = 0.0;
			for(j = 0;j<fDIndexToFDTransRew[i].length;j++){
				if(fDIndexToMaxFDTransRew[i]< fDIndexToFDTransRew[i][j])
					fDIndexToMaxFDTransRew[i]= fDIndexToFDTransRew[i][j];
			}
			
			fDIndexToMaxRateRew[i] = 0.0;
			for(int k : fDIndexToPotato.get(i)){
				if(fDIndexToMaxRateRew[i] < fDCTMCRewards.getStateReward(k))
					fDIndexToMaxRateRew[i]= fDCTMCRewards.getStateReward(k);
			}			
			
			fDIndexToMaxExpTransRew[i] = 0.0;
			for(j = 0;j<fDIndexToExpTransRew[i].length;j++){
				if(fDIndexToMaxExpTransRew[i]< fDIndexToExpTransRew[i][j])
					fDIndexToMaxExpTransRew[i]= fDIndexToExpTransRew[i][j];
			}

			
			
			//for computation of dMax and deltaPMin
			
			//first check wheher for thi potato applies dMax or deltaPMin
			//we perform if out states are reached with smaller probability than 1.0 from entry then dMax should be applied
			for(int k : fDIndexToPotato.get(i)){
				tempTarget2.set(k, false);
			}

			System.out.println(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>" + fDIndexToEntry[i]);
			System.out.println("tempTarget2" + tempTarget2 );

			//TODO the setting of values of init and known may cause some speedup, check it up!
			reachOutStates = dTMCModelChecker.computeReachProbs(uniformisedDTMC, null, tempTarget2, null, null).soln[fDIndexToEntry[i]];
			for(int k : fDIndexToPotato.get(i))
				tempTarget2.set(k, true);

			
			System.out.println("reachOutStates: " + reachOutStates);
			useDMax = reachOutStates < 1-1.0e-11; 
			if(useDMax){
				//use dMax for computation of maximum delay
				
				//we first compute the minimum state reward in potato
				minStateRew = Double.MAX_VALUE;
				//state rewards
				for(int k : fDIndexToPotato.get(i)) {
					if(minStateRew > fDCTMCRewards.getStateReward(k)&& fDCTMCRewards.getStateReward(k) > 0)
						minStateRew=fDCTMCRewards.getStateReward(k);
				}
				//now we compute the dMax
				dMax = maxValue/((1-reachOutStates)*minStateRew);
				
				System.out.println("dMax: " + dMax); //TODO erase
			} else {
				//use deltaPMin for computation of maximum delay 
				minDeltaProbOrRew = 1.0e-7; //TODO perform accurate computation!!!
			}
			

			System.out.println("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
			
			
			
			System.out.println("reachEntry: " + reachEntry);
			System.out.println("maxValue: " + maxValue);
			System.out.println("fDIndexToMaxFDTransRew: " + fDIndexToMaxFDTransRew[i]);
			System.out.println("maxRateRew: " + fDIndexToMaxRateRew[i]);
			System.out.println("maxExpTransRew: " + fDIndexToMaxExpTransRew[i]);

			
			
			//////////////////// discretize down
			
			currDelay = fDPositionToDelays.get(i).get(currentFDPositions[i]);
			nextDistr = currDistr = fDPositionToDistrs.get(i).get(currentFDPositions[i]); 
			//first we store to delta the probability to reach target from current entry without visiting entry state
			//TODO kod dole zlozity, dali by sa naviac k distribuciam ukladat aj polia- sice redundancia, ale bolo by to rychlejsie
			delta = 0.0;
			distrIt = nextDistr.iterator(); 
			while(distrIt.hasNext()){
				distrEntry = distrIt.next();
				delta += distrEntry.getValue() * reachProbs[distrEntry.getKey()];
			}
			delta = (delta * error) / (model.getNumFDEvents()*reachEntry*
					(fDIndexToMaxRateRew[i] + uniformizationRate*(maxValue + fDIndexToMaxFDTransRew[i] + fDIndexToMaxExpTransRew[i])));

			nextDelay = currDelay - delta;
			
			System.out.println(">>>>>>>>>>>>>>>>>>>>>> FD: " + i + " useDMax: " + useDMax);
			System.out.println(1+". DELAY: " + currDelay);
			while(nextDelay >0){
				//compute new distribution and reward
				nextDistr = processPotatoState(i, fDIndexToEntry[i], nextDelay,
						nextDelay, 1e-11, nextReward, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
				
				//store computed values
				fDpositionToRewards.get(i).add(nextReward[0]);
				fDPositionToDistrs.get(i).add(nextDistr);
				fDPositionToDelays.get(i).add(nextDelay);
				fDOrderToPosition.get(i).addFirst(fDPositionToDistrs.get(i).size()-1);
				
//				System.out.println(fDPositionToDistrs.get(i).size()+". DELAY: " + nextDelay + " delta: " + delta);
				//TODO ked je bound na pocet krokov vyssi ako bound spocitany predchadzajucou metodou pouzi ten bound :)
				
				if(fDPositionToDistrs.get(i).size() == 300) break;
				
				delta = 0.0;
				distrIt = nextDistr.iterator(); 
				while(distrIt.hasNext()){
					distrEntry = distrIt.next();
					delta += distrEntry.getValue() * reachProbs[distrEntry.getKey()];
				}

				delta = (delta * error) / (model.getNumFDEvents()*reachEntry*
						(fDIndexToMaxRateRew[i] + uniformizationRate*(maxValue + fDIndexToMaxFDTransRew[i] + fDIndexToMaxExpTransRew[i])));

				//compute new delay
				currDelay = nextDelay;
				currDistr = nextDistr;
				nextDelay = currDelay - delta;
				
//				System.out.println("STEP BOUND: " + (1/delta));				
			}
			System.out.println("DOWN number of points: " +fDPositionToDistrs.get(i).size());			

			
			//////////////////// discretize up
			
			currMaxDiff = 1.0;
			currDelay = fDPositionToDelays.get(i).get(currentFDPositions[i]);
			currReward = fDpositionToRewards.get(i).get(currentFDPositions[i]);
			nextDistr = currDistr = fDPositionToDistrs.get(i).get(currentFDPositions[i]);
			//first we store to delta the probability to reach target from current entry without visiting entry state
			//TODO kod dole zlozity, dali by sa naviac k distribuciam ukladat aj polia- sice redundancia, ale bolo by to rychlejsie
			delta = 0.0;
			distrIt = nextDistr.iterator(); 
			while(distrIt.hasNext()){
				distrEntry = distrIt.next();
				delta += distrEntry.getValue() * reachProbs[distrEntry.getKey()];
			}
			delta = (delta * error) / (model.getNumFDEvents()*reachEntry*
					(uniformizationRate*(maxValue + fDIndexToMaxFDTransRew[i])));

			nextDelay = currDelay + delta;
			
			System.out.println(">>>>>>>>>>>>>>>>>>>>>> FD: " + i);
			System.out.println(1+". DELAY: " + currDelay);
			while((useDMax && (nextDelay < dMax)) || (!useDMax && (currMaxDiff > minDeltaProbOrRew))){
				//compute new distribution and reward
				nextDistr = processPotatoState(i, fDIndexToEntry[i], nextDelay,
						nextDelay, 1e-11, nextReward, false); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
				
				//store computed values
				fDpositionToRewards.get(i).add(nextReward[0]);
				fDPositionToDistrs.get(i).add(nextDistr);
				fDPositionToDelays.get(i).add(nextDelay);
				fDOrderToPosition.get(i).addLast(fDPositionToDistrs.get(i).size()-1);
				
//				System.out.println(fDPositionToDistrs.get(i).size()+". DELAY: " + nextDelay + " delta: " + delta);
				//TODO ked je bound na pocet krokov vyssi ako bound spocitany predchadzajucou metodou pouzi ten bound :)
				
//				if(fDPositionToDistrs.get(i).size() == 5) break;
				

				delta = 0.0;
				distrIt = nextDistr.iterator(); 
				while(distrIt.hasNext()){
					distrEntry = distrIt.next();
					delta += distrEntry.getValue() * reachProbs[distrEntry.getKey()];
				}

				delta = (delta * error) / (model.getNumFDEvents()*reachEntry*
						(uniformizationRate*(maxValue + fDIndexToMaxFDTransRew[i])));
				
				if(!useDMax){
					currMaxDiff = 0.0;
					//compute max difference in distributions
					distrIt = nextDistr.iterator(); 
					while(distrIt.hasNext()){
						distrEntry = distrIt.next();
						currMaxDiff += distrEntry.getValue();
					}
					distrIt = currDistr.iterator(); 
					while(distrIt.hasNext()){
						distrEntry = distrIt.next();
						currMaxDiff -= distrEntry.getValue();
					}
					//compute max difference in one-step rewards
					if (currMaxDiff < (nextReward[0] - currReward)) 
						currMaxDiff = (nextReward[0] - currReward);
				}
				
				//compute new delay
				currReward = nextReward[0];
				currDelay  = nextDelay;
				currDistr  = nextDistr;
				nextDelay  = currDelay + delta;
				
//				System.out.println("currMaxDiff: " + currMaxDiff);
			}			
			System.out.println("UP number of points: " +fDPositionToDistrs.get(i).size() + " LAST DELAY: " + currDelay);

		}
		
		int minimalPosition, currentPosition, nextPosition;
		double minimalUpdate;
		double currentUpdate, currentDeltaRewBound, nextDeltaRewBound;
		boolean doNextStep = true;
		ListIterator<Integer> listIterator;

		for(j = 0;j<values.length; j++)
			System.out.print(values[j] + ", ");
		System.out.println("<<<<<<<<<<<<<<<<<<<<<<");
		
		//main cycle 
		while(doNextStep){
			//choose new actions and test whether are same as current actions
			//if they are the same, stop
			doNextStep = false;
			for(i =0; i<model.getNumFDEvents();i++){
				System.out.println(">>>>>>>>>>>>>>i: " + i);
				//we first compute the update for current choice
				//this will ensure that when, there are multiple best updates we will pick the same one as in the previous step 
				currentUpdate = fDpositionToRewards.get(i).get(currentFDPositions[i]);
				distrIt = fDPositionToDistrs.get(i).get(currentFDPositions[i]).iterator();
				while(distrIt.hasNext()){
					distrEntry = distrIt.next();
					currentUpdate += values[distrEntry.getKey()]*distrEntry.getValue();
				}
				minimalPosition = currentFDPositions[i];
				minimalUpdate = currentUpdate;
				for(int k: fDOrderToPosition.get(i)){
					currentUpdate = fDpositionToRewards.get(i).get(k);
					distrIt = fDPositionToDistrs.get(i).get(k).iterator();
					while(distrIt.hasNext()){
						distrEntry = distrIt.next();
						currentUpdate += values[distrEntry.getKey()]*distrEntry.getValue();
					}
					if(currentUpdate < minimalUpdate) {
						minimalUpdate = currentUpdate;
						minimalPosition = k;
					}
//					System.out.print(", " + minimalPosition);
				}
//				System.out.println(" ");
				
				if(currentFDPositions[i] != minimalPosition) {
					currentFDPositions[i] = minimalPosition;
					doNextStep = true;
				}
			}
			
			if(!doNextStep) break;
			
			//update values
			for(i=0;i<model.getNumFDEvents();i++){
				mDPState = fDCTMCStateToMDPState[fDIndexToEntry[i]];
				index = currentFDPositions[i];

				dTMC.setTransitions(mDPState, fDPositionToDistrs.get(i).get(index));
				dTMCRewards.setStateReward(mDPState, fDpositionToRewards.get(i).get(index));
			}
			
			//TODO the setting of values of init and known may cause some speedup, check it up!		
			values = dTMCModelChecker.computeReachRewards(dTMC, dTMCRewards, target, null, null).soln;

			for(j = 0;j<values.length; j++)
				System.out.print(values[j] + ", ");
			System.out.println("<<<<<<<<<<<<<<<<<<<<<<");
			
			//test discretization and add new points if needed
			for(i =0; i<model.getNumFDEvents();i++){
				mDPState = fDCTMCStateToMDPState[fDIndexToEntry[i]];
				
				//compute p(s)(t)
				//use method for computation of reachability probabilities in DTMC where there is self-loop with prob. 1 on s
				// the method should return vector of reachability probabilities of target states from any state
				remain.set(mDPState, false);
				//TODO the setting of values of init and known may cause some speedup, check it up!
				reachProbs = dTMCModelChecker.computeReachProbs(dTMC, remain, target, null, null).soln;
				remain.set(mDPState, true);

				reachEntry = 1-reachProbs[0];
				
				//discretization 
				//first compute necessary constants
				//compute fDIndexToMaxFDTransRew, maxRateRew and maxExpTransRew
				maxValue = 0.0;
				for(j = 0;j<values.length;j++){
					if(maxValue< values[j])
						maxValue= values[j];
				}
				
				//for computation of dMax and deltaPMin
				
				//first check wheher for thi potato applies dMax or deltaPMin
				//we perform if out states are reached with smaller probability than 1.0 from entry then dMax should be applied
				for(int k : fDIndexToPotato.get(i)){
					tempTarget2.set(k, false);
				}
				//TODO the setting of values of init and known may cause some speedup, check it up!
				reachOutStates = dTMCModelChecker.computeReachProbs(uniformisedDTMC, null, tempTarget2, null, null).soln[fDIndexToEntry[i]];
				for(int k : fDIndexToPotato.get(i))
					tempTarget2.set(k, true);

				
				System.out.println("reachOutStates: " + reachOutStates);
				useDMax = reachOutStates < 1-1.0e-11; 
				if(useDMax){
					//use dMax for computation of maximum delay
					
					//we first compute the minimum state reward in potato
					minStateRew = Double.MAX_VALUE;
					//state rewards
					for(int k : fDIndexToPotato.get(i)) {
						if(minStateRew > fDCTMCRewards.getStateReward(k)&& fDCTMCRewards.getStateReward(k) > 0)
							minStateRew=fDCTMCRewards.getStateReward(k);
					}
					//now we compute the dMax
					dMax = maxValue/((1-reachOutStates)*minStateRew);
					
					System.out.println("dMax: " + dMax); //TODO erase
				} else {
					//use deltaPMin for computation of maximum delay 
					minDeltaProbOrRew = 1.0e-7; //TODO perform accurate computation!!!
				}
				

				System.out.println("reachEntry: " + reachEntry);
				System.out.println("maxValue: " + maxValue);
				System.out.println("fDIndexToMaxFDTransRew: " + fDIndexToMaxFDTransRew[i]);
				System.out.println("maxRateRew: " + fDIndexToMaxRateRew[i]);
				System.out.println("maxExpTransRew: " + fDIndexToMaxExpTransRew[i]);

				
				
				
				//TODO test the elements around dMin 
			
				
				//check the elements right next to each other until  last element in fDOrderToPosition.get(i)
				//i.e. checking elements until old dMax
				listIterator = fDOrderToPosition.get(i).listIterator(1);
				
				currentPosition = fDOrderToPosition.get(i).get(0);
				currentDeltaRewBound = fDpositionToRewards.get(i).get(currentPosition);
				distrIt = fDPositionToDistrs.get(i).get(currentPosition).iterator();
				while(distrIt.hasNext()){
					distrEntry = distrIt.next();
					currentDeltaRewBound += values[distrEntry.getKey()]*distrEntry.getValue();
				}
				
				System.out.println("HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH");
				while(listIterator.hasNext()){
					nextPosition = listIterator.next();
					nextDeltaRewBound = fDpositionToRewards.get(i).get(nextPosition);
					distrIt = fDPositionToDistrs.get(i).get(nextPosition).iterator();
					delta = 0.0;
					while(distrIt.hasNext()){
						distrEntry = distrIt.next();
						nextDeltaRewBound += values[distrEntry.getKey()]*distrEntry.getValue();
						delta += distrEntry.getValue() * reachProbs[distrEntry.getKey()];
					}



					delta = Math.abs(model.getNumFDEvents()*reachEntry*(currentDeltaRewBound-nextDeltaRewBound)/delta); 
					if(error < delta) {
						System.out.print(",         NOK: " + currentPosition);
						//TODO code here discretize up from fDPositionToDelays.get(i).get(currentPosition)
						//until meeting fDPositionToDelays.get(i).get(nextPosition)
					}
					currentPosition = nextPosition;
					currentDeltaRewBound = nextDeltaRewBound;
					
					
//					//////////////////// discretize up
//					
//					currMaxDiff = 1.0;
//					currDelay = fDPositionToDelays.get(i).get(currentFDPositions[i]);
//					currReward = fDpositionToRewards.get(i).get(currentFDPositions[i]);
//					nextDistr = currDistr = fDPositionToDistrs.get(i).get(currentFDPositions[i]);
//					//first we store to delta the probability to reach target from current entry without visiting entry state
//					//TODO kod dole zlozity, dali by sa naviac k distribuciam ukladat aj polia- sice redundancia, ale bolo by to rychlejsie
//					delta = 0.0;
//					distrIt = nextDistr.iterator(); 
//					while(distrIt.hasNext()){
//						distrEntry = distrIt.next();
//						delta += distrEntry.getValue() * reachProbs[distrEntry.getKey()];
//					}
//					delta = (delta * error) / (model.getNumFDEvents()*reachEntry*
//							(uniformizationRate*(maxValue + fDIndexToMaxFDTransRew[i])));
//
//					nextDelay = currDelay + delta;
//					
//					System.out.println(">>>>>>>>>>>>>>>>>>>>>> FD: " + i);
//					System.out.println(1+". DELAY: " + currDelay);
//					while((useDMax && (nextDelay < dMax)) || (!useDMax && (currMaxDiff > minDeltaProbOrRew))){
//						//compute new distribution and reward
//						nextDistr = processPotatoState(i, fDIndexToEntry.get(i), nextDelay,
//								nextDelay, 1e-11, nextReward); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
//						
//						//store computed values
//						fDpositionToRewards.get(i).add(nextReward[0]);
//						fDPositionToDistrs.get(i).add(nextDistr);
//						fDPositionToDelays.get(i).add(nextDelay);
//						fDOrderToPosition.get(i).addLast(fDPositionToDistrs.get(i).size()-1);
//						
////						System.out.println(fDPositionToDistrs.get(i).size()+". DELAY: " + nextDelay + " delta: " + delta);
//						//TODO ked je bound na pocet krokov vyssi ako bound spocitany predchadzajucou metodou pouzi ten bound :)
//						
////						if(fDPositionToDistrs.get(i).size() == 5) break;
//						
//
//						delta = 0.0;
//						distrIt = nextDistr.iterator(); 
//						while(distrIt.hasNext()){
//							distrEntry = distrIt.next();
//							delta += distrEntry.getValue() * reachProbs[distrEntry.getKey()];
//						}
//
//						delta = (delta * error) / (model.getNumFDEvents()*reachEntry*
//								(uniformizationRate*(maxValue + fDIndexToMaxFDTransRew[i])));
//						
//						if(!useDMax){
//							currMaxDiff = 0.0;
//							//compute max difference in distributions
//							distrIt = nextDistr.iterator(); 
//							while(distrIt.hasNext()){
//								distrEntry = distrIt.next();
//								currMaxDiff += distrEntry.getValue();
//							}
//							distrIt = currDistr.iterator(); 
//							while(distrIt.hasNext()){
//								distrEntry = distrIt.next();
//								currMaxDiff -= distrEntry.getValue();
//							}
//							//compute max difference in one-step rewards
//							if (currMaxDiff < (nextReward[0] - currReward)) 
//								currMaxDiff = (nextReward[0] - currReward);
//						}
//						
//						//compute new delay
//						currReward = nextReward[0];
//						currDelay  = nextDelay;
//						currDistr  = nextDistr;
//						nextDelay  = currDelay + delta;
//						
////						System.out.println("currMaxDiff: " + currMaxDiff);
//					}
					
				}
				
				//TODO check the elements after dMax whether we don't have to move it up

				
//				//////////////////// discretize up
//				
//				currMaxDiff = 1.0;
//				currDelay = fDPositionToDelays.get(i).get(currentFDPositions[i]);
//				currReward = fDpositionToRewards.get(i).get(currentFDPositions[i]);
//				nextDistr = currDistr = fDPositionToDistrs.get(i).get(currentFDPositions[i]);
//				//first we store to delta the probability to reach target from current entry without visiting entry state
//				//TODO kod dole zlozity, dali by sa naviac k distribuciam ukladat aj polia- sice redundancia, ale bolo by to rychlejsie
//				delta = 0.0;
//				distrIt = nextDistr.iterator(); 
//				while(distrIt.hasNext()){
//					distrEntry = distrIt.next();
//					delta += distrEntry.getValue() * reachProbs[distrEntry.getKey()];
//				}
//				delta = (delta * error) / (model.getNumFDEvents()*reachEntry*
//						(uniformizationRate*(maxValue + fDIndexToMaxFDTransRew[i])));
//
//				nextDelay = currDelay + delta;
//				
//				System.out.println(">>>>>>>>>>>>>>>>>>>>>> FD: " + i);
//				System.out.println(1+". DELAY: " + currDelay);
//				while((useDMax && (nextDelay < dMax)) || (!useDMax && (currMaxDiff > minDeltaProbOrRew))){
//					//compute new distribution and reward
//					nextDistr = processPotatoState(i, fDIndexToEntry.get(i), nextDelay,
//							nextDelay, 1e-11, nextReward); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
//					
//					//store computed values
//					fDpositionToRewards.get(i).add(nextReward[0]);
//					fDPositionToDistrs.get(i).add(nextDistr);
//					fDPositionToDelays.get(i).add(nextDelay);
//					fDOrderToPosition.get(i).addLast(fDPositionToDistrs.get(i).size()-1);
//					
////					System.out.println(fDPositionToDistrs.get(i).size()+". DELAY: " + nextDelay + " delta: " + delta);
//					//TODO ked je bound na pocet krokov vyssi ako bound spocitany predchadzajucou metodou pouzi ten bound :)
//					
////					if(fDPositionToDistrs.get(i).size() == 5) break;
//					
//
//					delta = 0.0;
//					distrIt = nextDistr.iterator(); 
//					while(distrIt.hasNext()){
//						distrEntry = distrIt.next();
//						delta += distrEntry.getValue() * reachProbs[distrEntry.getKey()];
//					}
//
//					delta = (delta * error) / (model.getNumFDEvents()*reachEntry*
//							(uniformizationRate*(maxValue + fDIndexToMaxFDTransRew[i])));
//					
//					if(!useDMax){
//						currMaxDiff = 0.0;
//						//compute max difference in distributions
//						distrIt = nextDistr.iterator(); 
//						while(distrIt.hasNext()){
//							distrEntry = distrIt.next();
//							currMaxDiff += distrEntry.getValue();
//						}
//						distrIt = currDistr.iterator(); 
//						while(distrIt.hasNext()){
//							distrEntry = distrIt.next();
//							currMaxDiff -= distrEntry.getValue();
//						}
//						//compute max difference in one-step rewards
//						if (currMaxDiff < (nextReward[0] - currReward)) 
//							currMaxDiff = (nextReward[0] - currReward);
//					}
//					
//					//compute new delay
//					currReward = nextReward[0];
//					currDelay  = nextDelay;
//					currDistr  = nextDistr;
//					nextDelay  = currDelay + delta;
//					
////					System.out.println("currMaxDiff: " + currMaxDiff);
//				}

			}
			
		}
		return values;
	}
		
	public void discretizationMethod(double error) throws PrismException {

		MDPModelChecker mDPModelChecker = new MDPModelChecker(this); 
		ModelCheckerResult result = null;
		
		//we use initial values FDs to compute corresponding one-step rewards and transition probabilities
		//we fill corresponding MDP and solve it
		double[] temp = new double[1];
		for(int i=0;i<fDIndexToEntry.length;i++) {
			processPotatoState(i, fDIndexToEntry[i], model.getFDEvent(i).getDelayTime(),
					model.getFDEvent(i).getDelayTime(), 1e-11, temp, true); //TODO error should be computed accordingly, i.e. try some error, then test perturbation bound, if it is OK, let it be, otherwise make another iteration with smaller error
		}
	
		//model check created MDP
		
		mainLog.println("Created DTMC for computation of expected cost:\n" + mdp);
		mainLog.println("Created DTMC costs:\n" + mdpRewards+ "\n");
		mainLog.println("Created DTMC target states:\n" + target+ "\n");
		
		result = mDPModelChecker.computeReachRewards(mdp, mdpRewards, target, true, null, null);		
		
		//computation of relevant constants
		double boundExpectedCost;
		double boundNumOfSteps;
		double alpha;
		double d1;

				
		//////////// boundExpectedCost
		
		//we evaluate bound on expected cost from every state
		//take maximum from all MDP states
		boundExpectedCost = 0.0;
		for(int i=0; i<mdp.getNumStates();i++){
			if(boundExpectedCost < result.soln[i]) 
				boundExpectedCost = result.soln[i];
		}
		
		mainLog.println("\n=========================================\n");
		mainLog.println("BOUND ON EXPECTED COST FROM EACH STATE: " + boundExpectedCost);
		mainLog.println("EXPECTED COST FROM INITIAL STATE: " + result.soln[mdp.getFirstInitialState()]);

		///////// boundNumOfSteps
		
		//we check whether each entry of potato has FD transition with positive impulse cost 
		double minimalOneStepCost = Double.MAX_VALUE;
		for(int i = 0; i<mDPStateToFDCTMCState.length; i++){
			int state = mDPStateToFDCTMCState[i];
			
			if(targetStatesFDCTMC.contains(state)) continue; //skip target states

			double currentCost = 0.0;
			//check whether there is FD event and process accordingly
			if(!fDCTMCStateToFDIndex.containsKey(state)) { 
				//non FD state
//				//first compute impulse costs
//				Iterator<Entry<Integer,Double>> it2;
//				it2 = model.getTransitionsIterator(state);
//				while(it2.hasNext()) {
//					Entry<Integer,Double> entry2 =it2.next();
//					currentCost = currentCost
//										+ entry2.getValue() * fDCTMCRewards.getTransitionReward(-1,
//												state, entry2.getKey());
//	                        }
//				//add rate costs
//				currentCost = currentCost + fDCTMCRewards.getStateReward(state)/model.getExitRate(state);;
//
//				//process current minimal cost 
				if(minimalOneStepCost > mdpRewards.getTransitionReward(fDCTMCStateToMDPState[state], 0))
					minimalOneStepCost = mdpRewards.getTransitionReward(fDCTMCStateToMDPState[state], 0);
		
			} else {
				//there is active FD event
				
				//the result one step cost is minimum of FD impulse cost and
				//exp impulse cost and rate costs accumulated in potato
				
				//it is possible to traverse potato and compute the cost on exp trans accumulated in potato
				//for now much simpler and correct implementation (probably could be done better)

				Iterator<Integer> it3 = fDIndexToPotato.get(fDCTMCStateToFDIndex.get(state)).iterator();
				int state2;
				while(it3.hasNext()) {
					state2 = it3.next();

					//first process the impulse cost of FD event
					Iterator<Entry<Integer,Double>> it;
					it = model.getFDEvent(fDCTMCStateToFDIndex.get(state2)).getTransitionsIterator(state2);
					while(it.hasNext()) {
						Entry<Integer,Double> entry2 =it.next();
						currentCost = currentCost 
											+ entry2.getValue() * fDCTMCRewards.getTransitionReward(fDCTMCStateToFDIndex.get(state2), 
													state2 , entry2.getKey());
					}
					if(currentCost<=0)
						throw new PrismException("State " + state2 
								+ " is entry of potato but does not have positive cost (" + currentCost +") on FD transition!");
					if(minimalOneStepCost > currentCost )
						minimalOneStepCost = currentCost;
					
					currentCost = 0.0;
					//first compute impulse costs
					Iterator<Entry<Integer,Double>> it2;
					it2 = model.getTransitionsIterator(state2);
					while(it2.hasNext()) {
						Entry<Integer,Double> entry2 =it2.next();
						currentCost = currentCost
											+ entry2.getValue() * fDCTMCRewards.getTransitionReward(-1,
													state2, entry2.getKey());
		                        }
					//add rate costs
					currentCost = currentCost + fDCTMCRewards.getStateReward(state2)/uniformizationRates[fDCTMCStateToFDIndex.get(state)];
	
					//process current minimal cost 
					if(minimalOneStepCost > currentCost)
						minimalOneStepCost = currentCost;
				}
			}
		}			

//		mainLog.println("MINIMAL ONE STEP COST IS: " + minimalOneStepCost);
		
		boundNumOfSteps = boundExpectedCost/minimalOneStepCost;
	
		mainLog.println("BOUND ON EXPECTED NUMBER OF STEPS FROM EACH STATE: "+ boundNumOfSteps);
		
		
		/////////////// alpha
		
		alpha = error /(boundNumOfSteps * (1+boundExpectedCost*mdp.getNumStates()));
		double alpha2 = 1/(2*boundNumOfSteps*mdp.getNumStates());
		if(alpha2 < alpha ) alpha = alpha2;
		
		mainLog.println("ALPHA: " + alpha);

		mainLog.println("\n============ common bounds computed ==============\n");
		
		long timer;
		double maxStateReward, maxImpulseReward, minStateReward, minBranchingPst, kappa, dMax, dMax2,maxRateSum, maxRateSumTemp;
		double[] deltas= new double[model.getNumFDEvents()];
		int fDIndex;

		for(fDIndex=0;fDIndex<fDIndexToEntry.length;fDIndex++) {
			mainLog.println("Computation of potato for FD " + fDIndex);
			

			////////////// d1
			
			//computation of minimal and maximal nonzero costs/rewards
			maxStateReward = 0.0; // = 2;
			maxImpulseReward = 0.0; // = 2;
			minStateReward = Double.MAX_VALUE; // = 1;
			minBranchingPst = Double.MAX_VALUE; // = 0.5;
			maxRateSum = 0.0;
			
			//exponential transitions
			for(int i : fDIndexToPotato.get(fDIndex)) {
				Iterator<Entry<Integer,Double>> it;
				it = model.getTransitionsIterator(i);
				maxRateSumTemp = 0;
				while(it.hasNext()) {
					Entry<Integer,Double> entry2 =it.next();
					maxRateSumTemp += entry2.getValue(); 
					if(minBranchingPst > entry2.getValue() && entry2.getValue() > 0)
						minBranchingPst = entry2.getValue();
					if(maxImpulseReward < fDCTMCRewards.getTransitionReward(-1,i, entry2.getKey()))
						maxImpulseReward = fDCTMCRewards.getTransitionReward(-1,i, entry2.getKey());
				}
				if(maxRateSum < maxRateSumTemp) 
					maxRateSum = maxRateSumTemp;
			}

			minBranchingPst = minBranchingPst/maxRateSum;
			
			//FD transitions
			for(int i =0; i<model.getNumStates();i++) {
				Iterator<Entry<Integer,Double>> it;
				it = model.getFDEvent(fDIndex).getTransitionsIterator(i);
				while(it.hasNext()) {
					Entry<Integer,Double> entry2 =it.next();
					if(minBranchingPst > entry2.getValue() && entry2.getValue() > 0)
						minBranchingPst = entry2.getValue();
					if(maxImpulseReward < fDCTMCRewards.getTransitionReward(fDIndex,i, entry2.getKey()))
						maxImpulseReward = fDCTMCRewards.getTransitionReward(fDIndex,i, entry2.getKey());
				}
			}

			
			//state rewards
			for(int i =0; i<model.getNumStates();i++) { //TODO is it sufficient to test only states in potato???
				if(minStateReward > fDCTMCRewards.getStateReward(i)&& fDCTMCRewards.getStateReward(i) > 0)
					minStateReward=fDCTMCRewards.getStateReward(i);
				if(maxStateReward < fDCTMCRewards.getStateReward(i))
					maxStateReward = fDCTMCRewards.getStateReward(i);
			}
			
			
			mainLog.println("\n\tmaxStateReward: " + maxStateReward + "\n\tmaxImpulseReward: " + maxImpulseReward 
					+ "\n\tminStateReward: " + minStateReward + "\n\tminBranchingPst: "+ minBranchingPst);
			
			//end of computation of minimal and maximal nonzero costs/rewards
			
			
			d1 = uniformizationRates[fDIndex]* maxStateReward + 2*uniformizationRates[fDIndex] * maxImpulseReward;
			if (d1< 2* uniformizationRates[fDIndex]) d1 = 2* uniformizationRates[fDIndex];
			
			mainLog.println("D1: " + d1);
			
			//compute delta and dmax for every potato    		
			deltas[fDIndex] = alpha/d1;
			kappa = (error * deltas[fDIndex] * minStateReward)/(2* mdp.getNumStates()*(1+Math.pow(boundExpectedCost,2)));
			//TODO traverse potato and find the minimum probability that some BSCC is reached in potato 
			dMax  = boundExpectedCost/(Math.pow(minBranchingPst, fDIndexToPotato.get(fDIndex).size())*minStateReward);
			dMax2 = (Math.E * Math.abs(Math.log(alpha/2)))/(uniformizationRates[fDIndex]*minBranchingPst);
//			mainLog.println("DMAX: " + dMax + " \tDMaX2: " + dMax2);
			if(dMax2 > dMax) dMax = dMax2;

			
			   			
//    		delta =  0.009;//0.000009;    		
//    		kappa = 1e-9;
//    		dMax  = 24.8;
    		mainLog.println("\n\tDELTA: " + deltas[fDIndex] + "\n\tKAPPA: " + kappa + "\n\tD_MAX: "+ dMax + "\n");

    		mainLog.println("\n============ potato bounds computed ==============\n");
    		
    		double[] workaround = new double[1]; //TODO erase when final version will be known
			processPotatoState(fDIndex, fDIndexToEntry[fDIndex], deltas[fDIndex], dMax, kappa,workaround, true);
		} //done computation of this potato
		

//		mainLog.println("\nBUILT MDP "+ mdp);
//		mainLog.println("BUILT MDPRewards: "+ mdpRewards + "\n");

		mainLog.println("\n=========Computing optimal strategy for created MDP==========\n");
		timer = System.currentTimeMillis();
		
		//model check created MDP
		mDPModelChecker.setGenStrat(true);
		result = mDPModelChecker.computeReachRewards(mdp, mdpRewards, target, true, null, null);

		timer =System.currentTimeMillis() - timer;
		mainLog.println("\nOptimal strategy computed in " + (timer/1000.0) + " s.\n");

		mainLog.print("\nRESULT in initial state: " + result.soln[mdp.getFirstInitialState()] + " delays: ");
		for(int i=0;i<fDIndexToEntry.length;i++) {
			mainLog.print(model.getFDEvent(i).getLabel() + "=" 
					+ ((((MDStrategy)result.strat).getChoiceIndex(fDCTMCStateToMDPState[fDIndexToEntry[i]])+1)*deltas[i]) + " ");
		}
		mainLog.println("");
	}
	
	
	
	/** 
	 * checking whether target states are in potato
	 * if yes we move it out of potato
	 */
	public void moveTargetStatesOutOfPotato() {
		FDEvent fD;
	    for(int j=0; j<model.getNumFDEvents();++j) {
	    	fD = model.getFDEvent(j); 
	    	for(int i=0;i<model.getNumStates();++i){
	    		if (fD.isActive(i) && targetStatesFDCTMC.contains(i)){
	    			fD.setPassive(i);
	    		}
	    	}
	    }
	    
	    //setting all target states as absorbing
	    Iterator<Integer> it;
	    int state;
	    it = targetStatesFDCTMC.iterator();
	    while(it.hasNext()) {
	    	state = it.next();
	    	model.clearState(state);
	    }
	    
	    double rate = model.getMaxExitRate();
	    
	    it = targetStatesFDCTMC.iterator();
	    while(it.hasNext()) {
	    	state = it.next();
	    	model.addToProbability(state, state, rate);
	    }
	}
	
	/**
	 * Creates mdp and mdpRewards. Fills states to mdp.
	 * Creates and fully fills all remaining global variables.
	 * @throws PrismException 
	 */
	private void initialise() throws PrismException{
		
		////////////////////checking whether target states are in potato
		//if yes we move it out of potato
		
		moveTargetStatesOutOfPotato();

		mainLog.println("CHANGED MODEL " + model);
		
		////////////////////creation of MDPs
		mdp = new MDPSimple();
        mdp.allowDupes = true;
		dTMC = new DTMCSimple();
        
        //find potatoes
        
        fDCTMCStateToFDIndex = new HashMap<Integer,Integer>(model.getNumStates());
		fDIndexToPotato = new ArrayList<Set<Integer>>(model.getNumFDEvents()); //FD to potato 
		fDIndexToOutStates = new ArrayList<Set<Integer>>(model.getNumFDEvents());

		
		Set<Integer> potato;
		FDEvent fD;
        for(int j=0; j<model.getNumFDEvents();++j) {
        	fD = model.getFDEvent(j); 
        	potato = new HashSet<Integer>();
        	for(int i=0;i<model.getNumStates();++i){
        		if (fD.isActive(i)){
        			if(fDCTMCStateToFDIndex.get(i) != null && fDCTMCStateToFDIndex.get(i) != fD.getIndex() )
        				throw new PrismException("Two active FDs in state " + i + "! Not supported.");
        			fDCTMCStateToFDIndex.put(i, j);
        			potato.add(i);
        		}
        	}
        	fDIndexToPotato.add(j, potato);
        	fDIndexToOutStates.add(j, new HashSet<Integer>());
        }
        
        mainLog.println("\n\nfDCTMCStateToFDIndex: " + fDCTMCStateToFDIndex);
        
        
        //fill necessary structures
		fDIndexToEntry = new int[model.getNumFDEvents()]; //FD to entry
		for(int i =0;i<fDIndexToEntry.length;i++) {
			fDIndexToEntry[i]=-1;
		}
		fDCTMCStateToMDPState = new int[model.getNumStates()];
		mDPStateToFDCTMCState = new int[model.getNumStates()];

		
		
		if(model.getNumInitialStates() != 1)
			mainLog.println("WARNING: detected multiple initial states. Using the first one.");


		//add newly found state
		int initialState = model.getFirstInitialState();
		int newMDPState = mdp.addState();
		mdp.addInitialState(newMDPState);
		dTMC.addState();
		dTMC.addInitialState(newMDPState);
		fDCTMCStateToMDPState[initialState] = newMDPState;
		mDPStateToFDCTMCState[newMDPState] = initialState;
		
		//process initial state
		if(fDCTMCStateToFDIndex.get(initialState)!=null){
			//set entry of potato to corresponding FD
			fDIndexToEntry[fDCTMCStateToFDIndex.get(initialState)] = initialState;
		}
			
		//structures for traversing
		Stack<Integer> toDoStates = new Stack<Integer>(); //grey states
		Set<Integer> doneStates = new HashSet<Integer>(model.getNumStates()); //black states
		Integer currentState, nextState;
		
		toDoStates.push(model.getFirstInitialState());
		
		while (!toDoStates.isEmpty()) {
			currentState = toDoStates.pop();
			doneStates.add(currentState);
			
			//explore exponential successors 
			Iterator<Entry<Integer,Double>> it = model.getTransitions(currentState).iterator(); //iterate over distribution put to stack, check conditions!!!
			while (it.hasNext()) {
				nextState = ((Entry<Integer,Double>)it.next()).getKey();
				
				if(!(doneStates.contains(nextState) || toDoStates.contains(nextState))){ //not found yet
				
					//check whether there is an active FD event
					if(fDCTMCStateToFDIndex.get(nextState) == null){ //there is none, it is an MDP state
			    		newMDPState = mdp.addState();
			    		dTMC.addState();
						fDCTMCStateToMDPState[nextState] = newMDPState;
						mDPStateToFDCTMCState[newMDPState] = nextState;    						
					}
				}
				
				//check whether this transition goes to entry of potato
				if(fDCTMCStateToFDIndex.get(nextState) != null &&  				//some FD must be active in nextState
						(fDCTMCStateToFDIndex.get(currentState) == null 		//and was not active before
																				//or before was active different FD
						  || (fDCTMCStateToFDIndex.get(currentState) != null 
						  		&& fDCTMCStateToFDIndex.get(currentState)!=fDCTMCStateToFDIndex.get(nextState))))		
				{
					if(!(doneStates.contains(nextState) || toDoStates.contains(nextState))){ //not found yet
			    		newMDPState = mdp.addState();
			    		dTMC.addState();
						fDCTMCStateToMDPState[nextState] = newMDPState;
						mDPStateToFDCTMCState[newMDPState] = nextState;
					}
					
					//check whether there is different entry to potato
					if(fDIndexToEntry[fDCTMCStateToFDIndex.get(nextState)] != -1 
							&& fDIndexToEntry[fDCTMCStateToFDIndex.get(nextState)] != nextState)
						throw new PrismException("There are multiple entries to potato! States: " + nextState
								+ " (reached from " + currentState
								+ ") and " + fDIndexToEntry[fDCTMCStateToFDIndex.get(nextState)]);
					fDIndexToEntry[fDCTMCStateToFDIndex.get(nextState)] = nextState;
				}
				
				//check whether transition leaves potato
				if(fDCTMCStateToFDIndex.get(currentState) != null &&  			//some FD must be active in currentState
						(fDCTMCStateToFDIndex.get(nextState) == null 			//and was not active in nextState
//						  || (targetStatesFDCTMC.contains(nextState))			//or transition leads to the target state
																				//or different FD is active in nextState
						  || (fDCTMCStateToFDIndex.get(nextState) != null 
						  		&& fDCTMCStateToFDIndex.get(currentState)!=fDCTMCStateToFDIndex.get(nextState))))
					fDIndexToOutStates.get(fDCTMCStateToFDIndex.get(currentState)).add(nextState);
				
				if(!(doneStates.contains(nextState) || toDoStates.contains(nextState))){ //not found yet
					toDoStates.push(nextState); //process traversing information
				}
			}
			
			//explore FD successors
			if(fDCTMCStateToFDIndex.get(currentState) == null) continue; //no FD
			it = model.getFDEvent(fDCTMCStateToFDIndex.get(currentState)).getTransitionsIterator(currentState);
			while (it.hasNext()) {
				nextState = ((Entry<Integer,Double>)it.next()).getKey();
				if(!(doneStates.contains(nextState) || toDoStates.contains(nextState))){ //not found yet

					//check whether there is an active FD event
					if(fDCTMCStateToFDIndex.get(nextState) == null){ //there is none, it is an MDP state
			    		newMDPState = mdp.addState();
			    		dTMC.addState();
						fDCTMCStateToMDPState[nextState] = newMDPState;
						mDPStateToFDCTMCState[newMDPState] = nextState;    						
					}
				}
				
				//check whether this transition goes to entry of potato
				if(fDCTMCStateToFDIndex.get(nextState) != null) {
					if(!(doneStates.contains(nextState) || toDoStates.contains(nextState))){ //not found yet
			    		newMDPState = mdp.addState();
			    		dTMC.addState();
						fDCTMCStateToMDPState[nextState] = newMDPState;
						mDPStateToFDCTMCState[newMDPState] = nextState;
					}
					
					//check whether there is already different entry to potato
					if(fDIndexToEntry[fDCTMCStateToFDIndex.get(nextState)] != -1 
							&& fDIndexToEntry[fDCTMCStateToFDIndex.get(nextState)] != nextState)
						throw new PrismException("There are multiple entries to potato! States: " + nextState
								+ " (reached from " + currentState
								+ ") and " + fDIndexToEntry[fDCTMCStateToFDIndex.get(nextState)]);
					
					fDIndexToEntry[fDCTMCStateToFDIndex.get(nextState)] = nextState;
				}
				
				//this transition leaves potato because it is FD
				fDIndexToOutStates.get(fDCTMCStateToFDIndex.get(currentState)).add(nextState);
				
				if(!(doneStates.contains(nextState) || toDoStates.contains(nextState))){ //not found yet
					toDoStates.push(nextState); //process traversing information
				}
			}
		}
		//done traversing

		//resize the mDPStateToFDCTMCState to the correct size
		int[] temp = new int[mdp.getNumStates()];
		for(int i=0;i<mdp.getNumStates();i++) {
			temp[i] = mDPStateToFDCTMCState[i];
		}
		mDPStateToFDCTMCState = temp;
		
		//use structures created when traversing to eliminate states from potato 
		//that are not used because they are unreachable i.e. thanks to the target states
		for(int i= 0; i<fDIndexToPotato.size();i++) {
			Iterator<Integer> it2 =  fDIndexToPotato.get(i).iterator(); 
			while(it2.hasNext()) {
				int potatoState = it2.next(); 
				if(!doneStates.contains(potatoState))
					it2.remove();
			}
		}
		
        mainLog.println("fDIndexToPotato: " + fDIndexToPotato);

		mainLog.println("fDIndexToEntry: "+ fDIndexToEntry);
		mainLog.println("fDCTMCStateToMDPState: " + fDCTMCStateToMDPState);
		mainLog.println("mDPStateToFDCTMCState: " + mDPStateToFDCTMCState);
		mainLog.println("fDIndexToOutStates: "+ fDIndexToOutStates);

		mdpRewards = new MDPRewardsSimple(mdp.getNumStates());
		dTMCRewards = new StateRewardsArray(mdp.getNumStates());
		
		
		fDIndexToFDTransRew  = new double[fDIndexToPotato.size()][];
		fDIndexToExpTransRew = new double[fDIndexToPotato.size()][];
		fDIndexToRateRew     = new double[fDIndexToPotato.size()][];

		//exp transitions
		fDIndexToExpToStates = new int[fDIndexToPotato.size()][][];
		fDIndexToExpToProbs  = new double[fDIndexToPotato.size()][][];
		//FD transitions
		fDIndexToFDToStates  = new int[fDIndexToPotato.size()][][];
		fDIndexToFDToProbs   = new double[fDIndexToPotato.size()][][];

		fDIndexToDTMCStatesToFDCTMCState = new int[fDIndexToPotato.size()][];
				
		uniformizationRates = new double[model.getNumFDEvents()];
		
		fDIndexToProbReachOutFromEntryThroughExp = new double[model.getNumFDEvents()];
	}
	
	/**
	 * Adds one choice to every non-potato state to mdp
	 * and corresponding rewards to mdpRewards.
	 */
    private void processNonPotatoStates() {
		//choices corresponding to non-entry to potato estates
		
		Iterator<Entry<Integer,Double>> it;
		double overallRate, overallCost;
		int src;
		for(int i=0;i<mdp.getNumStates();++i) {
			src = mDPStateToFDCTMCState[i];
			if(fDCTMCStateToFDIndex.containsKey(src)) continue; //entry-of potato -> continue

			overallRate = 0.0;
			it = model.getTransitionsIterator(src);			
			while (it.hasNext()) {
				Entry<Integer,Double> entry = ((Entry<Integer,Double>)it.next());
				overallRate += entry.getValue();
			}
			
			overallCost = (1/model.getExitRate(src)) * fDCTMCRewards.getStateReward(src); //add rate cost
			
			it = model.getTransitionsIterator(src);
			Distribution distr = new Distribution();
			while (it.hasNext()) {
				Entry<Integer,Double> entry = ((Entry<Integer,Double>)it.next());

				distr.add(fDCTMCStateToMDPState[entry.getKey()], entry.getValue()/overallRate);
				// add transition cost, i.e. trans_probability * impulse_cost
				overallCost += entry.getValue() * fDCTMCRewards.getTransitionReward(-1, src, entry.getKey());
			}
			
			
			
			mdp.addChoice(i, distr);
			dTMC.setTransitions(i, distr);
			mdpRewards.addToTransitionReward(i, 0, overallCost);
			dTMCRewards.setStateReward(i, overallCost);
		}
	}
    
	public void initialisePotatoState(int fDIndex, int entryToPotato) throws PrismException {
	
//		mainLog.println("FD INDEX: "+ fdIndex + " entry: " + entryToPotato);
		

		//initialization
		
		int dTMCInitialState = -1;
		int lastIndex, indexOfFirstNonPotatoState, numOfStates;
		numOfStates = fDIndexToPotato.get(fDIndex).size() + fDIndexToOutStates.get(fDIndex).size();
	
		
		DTMCSimple expDTMC, fDDTMC;
		expDTMC = new DTMCSimple();
		fDDTMC = new DTMCSimple();
		
		double[] fDTransRew, expTransRew, rateRew;
		fDTransRew = new double[numOfStates];
		expTransRew = new double[numOfStates];
		rateRew = new double[numOfStates];

	
		int[] dTMCStatesToFDCTMCState = new int[numOfStates];
		Map<Integer,Integer> fDCTMCStateToDTMCState = new HashMap<Integer,Integer>(fDIndexToOutStates.get(fDIndex).size());
		
		//add states
		//first add initial state
		lastIndex = expDTMC.addState();		
		fDDTMC.addState();
		dTMCStatesToFDCTMCState[lastIndex] = entryToPotato;
		dTMCInitialState = lastIndex;
		
		rateRew[lastIndex] = fDCTMCRewards.getStateReward(entryToPotato);

		if(lastIndex != 0)
			throw new PrismException("Index of entry state in the subordinated MC should be 0, but it is " + lastIndex + "!");
		
		//add states in potato
		for(int state: fDIndexToPotato.get(fDIndex)){ //TODO erase DTMCs sometime, they are already useless, but might be useful in the future
			if(state == entryToPotato) continue;
			lastIndex = expDTMC.addState();
			fDDTMC.addState();
			dTMCStatesToFDCTMCState[lastIndex] = state;
			fDCTMCStateToDTMCState.put(state, lastIndex);
		
			rateRew[lastIndex] = fDCTMCRewards.getStateReward(state);
		}
		indexOfFirstNonPotatoState = expDTMC.getNumStates();
		
		//add out states
		for(int state: fDIndexToOutStates.get(fDIndex)){
			lastIndex = expDTMC.addState();
			fDDTMC.addState();
			dTMCStatesToFDCTMCState[lastIndex] = state;
			fDCTMCStateToDTMCState.put(state, lastIndex);

			rateRew[lastIndex] = 0.0;
		}
		
	
		mainLog.println("fDCTMCStateToDTMCState: " + fDCTMCStateToDTMCState);

		
	
		//add exp transitions
		//first eliminate self loops and compute maximum exit rate
		Iterator<Entry<Integer,Double>> it;
		Map<Integer,Double> exitRates = new HashMap<Integer,Double>();
		double currentRate, maxRate = 0.0;
		for(int state: fDIndexToPotato.get(fDIndex)){
			currentRate = 0.0;
			it = model.getTransitionsIterator(state);
			while(it.hasNext()) {
				Entry<Integer,Double> entry2 = ((Entry<Integer,Double>)it.next());
				if (entry2.getKey() == state){
					//fix state rewards 
					fDCTMCRewards.addToStateReward(state, fDCTMCRewards.getTransitionReward(state, state) * entry2.getValue());
					rateRew[fDCTMCStateToDTMCState.get(state)] += fDCTMCRewards.getTransitionReward(state, state) * entry2.getValue();
					//remove self loop
					it.remove();
				} else {
					currentRate += entry2.getValue();
				}
			}
			exitRates.put(state, currentRate);
			if(maxRate < currentRate)
				maxRate = currentRate;
		}
		
		uniformizationRates[fDIndex]= maxRate;
		
		//now compute and add transition probabilities
		int[][] expToStates = new int[indexOfFirstNonPotatoState][];
		double[][] expToProbs = new double[indexOfFirstNonPotatoState][];
		int[] states;
		double[] probs;
		int source,dest = -1;
		for(int state: fDIndexToPotato.get(fDIndex)){
			if (state == entryToPotato) {
				source = dTMCInitialState;
			} else {
				source = fDCTMCStateToDTMCState.get(state);
			}

			if(exitRates.get(state) >= maxRate) {
				states = new int[model.getTransitions(state).size()]; 
				probs  = new double[model.getTransitions(state).size()];
			} else { //leaving space for self loop
				states = new int[model.getTransitions(state).size() + 1]; 
				probs  = new double[model.getTransitions(state).size() + 1];
				//add uniformization self loop
				states[model.getTransitions(state).size()] = source;
				probs[model.getTransitions(state).size()] = 1-(exitRates.get(state)/maxRate);

				expDTMC.addToProbability(source, source, 1-(exitRates.get(state)/maxRate));
			}
			
			it = model.getTransitionsIterator(state);
			double reward = 0.0;
			int j =0;
			while(it.hasNext()) {
				Entry<Integer,Double> entry2 = ((Entry<Integer,Double>)it.next());
				if (entry2.getKey() == entryToPotato) {
					dest = dTMCInitialState;
				} else {
					dest = fDCTMCStateToDTMCState.get(entry2.getKey());
				}
				
				states[j] = dest;
				probs[j] = entry2.getValue()/maxRate;
				j++;
				
				expDTMC.addToProbability(source, dest, entry2.getValue()/maxRate);
				
				//if trans reward is positive add it to reward vector 
				if(fDCTMCRewards.getTransitionReward(-1,state,entry2.getKey()) > 0){
					reward = reward + entry2.getValue() * fDCTMCRewards.getTransitionReward(-1,state,entry2.getKey());
				}
			}
			
			expToStates[source] = states;
			expToProbs[source]  = probs; 
			
			expTransRew[source] = reward;
		}
	
		//adding self loops to out states
		for(int i = indexOfFirstNonPotatoState;i < expDTMC.getNumStates();i++){
			expDTMC.addToProbability(i,i,1.0);
		}
		
		//add FD transitions
		int[][] fDToStates = new int[indexOfFirstNonPotatoState][];
		double[][] fDToProbs = new double[indexOfFirstNonPotatoState][];
		for(int state: fDIndexToPotato.get(fDIndex)){
			if (state == entryToPotato) {
				source = dTMCInitialState;
			} else {
				source = fDCTMCStateToDTMCState.get(state);
			}
			
			states = new int[model.getFDEvent(fDIndex).getTransitions(state).size()];
			probs  = new double[model.getFDEvent(fDIndex).getTransitions(state).size()];
			
			it = model.getFDEvent(fDIndex).getTransitionsIterator(state);
			double reward = 0.0;
			int j =0;
			while(it.hasNext()) {
				Entry<Integer,Double> entry2 = ((Entry<Integer,Double>)it.next());
				
				states[j] = fDCTMCStateToDTMCState.get(entry2.getKey());
				probs[j] = entry2.getValue();
				j++;
				
				fDDTMC.addToProbability(source,
						fDCTMCStateToDTMCState.get(entry2.getKey()),
						entry2.getValue());
				//if trans reward is positive add it to reward kernel 
				if(fDCTMCRewards.getTransitionReward(fDIndex,state,entry2.getKey()) > 0){
					reward = reward + entry2.getValue() * fDCTMCRewards.getTransitionReward(fDIndex,state,entry2.getKey());
				}
			}
			fDToStates[source] = states;
			fDToProbs[source]  = probs; 
			
			fDTransRew[source] = reward;
		}
		
		//storing computed values for future use
		fDIndexToFDTransRew[fDIndex]  = fDTransRew;
		fDIndexToExpTransRew[fDIndex] = expTransRew;
		fDIndexToRateRew[fDIndex]     = rateRew;

		fDIndexToExpToStates[fDIndex] = expToStates;
		fDIndexToExpToProbs[fDIndex]  = expToProbs;
		fDIndexToFDToStates[fDIndex]  = fDToStates;
		fDIndexToFDToProbs[fDIndex]   = fDToProbs;		
		
		fDIndexToDTMCStatesToFDCTMCState[fDIndex] = dTMCStatesToFDCTMCState; 
		
		
		////////// computing fDIndexToProbReachOutFromEntryThroughExp
		DTMCModelChecker dTMCModelChecker = new DTMCModelChecker(this); 
		BitSet tempTarget = new BitSet(model.getNumStates());
		for (int j=0;j<model.getNumStates();j++) {
			tempTarget.set(j, false);
		}
		//compute the probability of leaving potato by exponential transitions
		for(int state : fDIndexToOutStates.get(fDIndex))
			tempTarget.set(fDCTMCStateToDTMCState.get(state), true);
		//TODO the setting of values of init and known may cause some speedup, check it up!
		fDIndexToProbReachOutFromEntryThroughExp[fDIndex] = dTMCModelChecker.computeReachProbs(expDTMC, 
				null, tempTarget, null, null).soln[dTMCInitialState];
		///////// end of computation fDIndexToProbReachOutFromEntryThroughExp
		
		
		mainLog.println("\n expDTMC: "+ expDTMC);
		mainLog.println("\n fDDTMC: "+ fDDTMC);

		mainLog.println("\n fDIndexToProbReachOutFromEntryThroughExp: " + fDIndexToProbReachOutFromEntryThroughExp[fDIndex]);
		
		mainLog.print("\n expTransRew [");
		for(int i=0;i<(expTransRew.length-1);i++)
			mainLog.print(expTransRew[i] + ", ");
		mainLog.println(expTransRew[expTransRew.length-1] + "]");
		
		mainLog.print("\n fDTransRew [");
		for(int i=0;i<(fDTransRew.length-1);i++)
			mainLog.print(fDTransRew[i] + ", ");
		mainLog.println(fDTransRew[fDTransRew.length-1] + "]");

		mainLog.print("\n rateRew [");
		for(int i=0;i<(rateRew.length-1);i++)
			mainLog.print(rateRew[i] + ", ");
		mainLog.println(rateRew[rateRew.length-1] + "]");
	}

	
	
	public Distribution processPotatoState(int fDIndex, int entryToPotato, double delta, double dMax, double kappa, double reward[], boolean storeChoices) throws PrismException {

		//clear choices/rewards for this state in mdp/mdpRewards (new values will be computed)
		mdp.clearChoices(fDCTMCStateToMDPState[entryToPotato]);
		mdpRewards.clearRewards(fDCTMCStateToMDPState[entryToPotato]);

		double[] fDTransRew, expTransRew, rateRew;
		fDTransRew = fDIndexToFDTransRew[fDIndex];
		expTransRew = fDIndexToExpTransRew[fDIndex];
		rateRew = new double[fDIndexToRateRew[fDIndex].length];
		for(int i=0;i<fDIndexToRateRew[fDIndex].length;i++) {
			rateRew[i] = fDIndexToRateRew[fDIndex][i]*delta; //multiplied by delta in order not to do it in every step below;			
		}
		
		//exp transitions
		int[][] expToStates = fDIndexToExpToStates[fDIndex];
		double[][] expToProbs = fDIndexToExpToProbs[fDIndex];
		//FD transitions
		int[][] fDToStates = fDIndexToFDToStates[fDIndex];
		double[][] fDToProbs = fDIndexToFDToProbs[fDIndex];
		
		int[] dTMCStatesToFDCTMCState = fDIndexToDTMCStatesToFDCTMCState[fDIndex];
		
		int[] states;
		double[] probs;

		int numOfStates = fDTransRew.length;
		int indexOfFirstNonPotatoState = expToStates.length; 
		int numOfPoissonSteps = CbpAlg.getNumberOfPoissonStepsFromError(delta, ((Double)(dMax/delta)).intValue(), kappa, uniformizationRates[fDIndex]);
			
//		mainLog.println("NUM of poiss steps: " + numOfPoissonSteps);

			
		//////////////////computing and filling rate rewards

		long timer = System.currentTimeMillis();
		
		/////////////////computing transitions and rewards
		
		double[] temp;
		
		double[] transProb = new double[numOfStates];
		double[] transProb2 = new double[numOfStates];
		double[] resultProb = new double[numOfStates];
		
		//setting initial distribution
		// Initialise to 0
		for (int j = 0; j < numOfStates; j++) {
			transProb[j] = 0;
		}
		transProb[0]= 1.0; //assumption that entry of potato has index 0
		
		double exponential = Math.exp(-uniformizationRates[fDIndex] * delta);
	
//		mainLog.println("RATE: " + uniformizationRate + " DELTA: " + delta + " EXPONENTIAL: " + exponential);
		
		double overallExpTransReward;
		double weightedExpTransReward = 0.0;
		double overallRateReward;
		double weightedRateReward = 0.0;
		double lambdaT;
		int i, j, k, l, lastNonZeroIndex;

		int maxIters = (int) (dMax/delta);
		
		double sum, notAddedValue, prob, fixedDelayReward;
		fixedDelayReward = 0.0;
		Distribution distr = null;

		
//		mainLog.println("\nComputation of transition probabilities and transition costs.\n");
		timer = System.currentTimeMillis();

		//for each interval do
		for(i=1;i * delta <= dMax;++i){
			lambdaT = 1;

			overallRateReward = 0.0;
			
			//zero step
			//add to resultProb
			for(k = 0; k< numOfStates;k++)
				resultProb[k] = transProb[k] * exponential * lambdaT;
			
			//compute rate rewards
			for(k= 0; k< numOfStates;k++)
				overallRateReward = overallRateReward + rateRew[k]*transProb[k];
			weightedRateReward = weightedRateReward + overallRateReward * exponential * lambdaT; 
			
			overallExpTransReward = 0.0;
			for(j=1;j<=numOfPoissonSteps;++j) {
				lambdaT = lambdaT * (uniformizationRates[fDIndex] * delta /j); 
				
				//first compute exp. transition rewards
				for(k = 0; k< numOfStates;k++)
					overallExpTransReward = overallExpTransReward + expTransRew[k]* transProb[k];
				weightedExpTransReward = weightedExpTransReward + overallExpTransReward * exponential * lambdaT;
				
//				expDTMC.vmMult(transProb, transProb2);

				///////// begin of explicit vmMult
				// Initialise result to 0
				for (k = 0; k < indexOfFirstNonPotatoState; k++) {
					transProb2[k] = 0;
				} 
				// and assign same value to absorbing states
				for (k = indexOfFirstNonPotatoState; k < numOfStates; k++) {
					transProb2[k] = transProb[k];
				}
				// Go through matrix elements (by row)
				for (k = 0; k < indexOfFirstNonPotatoState; k++) {
					states = expToStates[k];
					probs = expToProbs[k];
					for (l=0;l < states.length;l++) {
						transProb2[states[l]] += probs[l] * transProb[k];
					}
				}
				///////// end of explicit vmMult 
				
				//add to resultProb
				for(k = 0; k< numOfStates;k++)
					resultProb[k] = resultProb[k] + (transProb2[k] * exponential * lambdaT);
	
				//compute rate rewards
				for(k= 0; k< numOfStates;k++)
					overallRateReward = overallRateReward + rateRew[k]*transProb2[k];
				weightedRateReward = weightedRateReward + overallRateReward * exponential * lambdaT/(j+1);
				
//				mainLog.println("J: " + j + " weighted reward: " + weightedRateReward 
//						+ " increase: " + (overallRateReward * exponential * lambdaT/(j+1)) 
//						+ " poisson: " + (exponential * lambdaT));
				
				//swap vectors
				temp = transProb2;
				transProb2 = transProb;
				transProb = temp;
	
			}

			
			//filling rewards to MDP

//			mainLog.println("RATE REWARD: " + weightedRateReward);
			
			//compute fixed delay rewards
			fixedDelayReward = 0.0;
			
			for(k = 0; k< numOfStates;k++)
				fixedDelayReward = fixedDelayReward + resultProb[k]* fDTransRew[k];

			//TODO allocates more memory and then it releases
			if(storeChoices)
				mdpRewards.setTransitionReward(fDCTMCStateToMDPState[entryToPotato], 
						i-1, fixedDelayReward + weightedExpTransReward + weightedRateReward);
			
//			mainLog.println("TRANSITION COSTS: " + (fixedDelayReward + weightedExpTransReward));

			//fire fixed delays 
//			fDDTMC.vmMult(resultProb, transProb2);
			///////// begin of explicit vmMult
			// Initialise result to 0
			for (k = 0; k < numOfStates; k++) {
				transProb2[k] = 0;
			}
			// Go through matrix elements (by row)
			for (k = 0; k < indexOfFirstNonPotatoState; k++) {
				states = fDToStates[k];
				probs = fDToProbs[k];
				for (l=0;l < states.length;l++) {
					transProb2[states[l]] += probs[l] * resultProb[k];
				}
			}
			///////// end of explicit vmMult 

			//TODO allocates more memory and then it releases
			//filling new choice for MDP
			distr = new Distribution();
			
			//copy probability from non-potato states to mdp
			//We should ensure that distribution sums to exactly one. For the last state
			//that has nonzero probability we set its probability to 1 - sum of the rest.
			sum =0.0;
			notAddedValue = 0;
			lastNonZeroIndex=-1;
			prob = 0.0;
			for(j=indexOfFirstNonPotatoState;j<numOfStates; ++j) {
				prob = resultProb[j]+transProb2[j];
				if(prob > 0){
					if(lastNonZeroIndex != -1){
						distr.set(fDCTMCStateToMDPState[dTMCStatesToFDCTMCState[lastNonZeroIndex]], notAddedValue);
						sum = sum + notAddedValue;
					}
					lastNonZeroIndex = j;
					notAddedValue = prob;
				}
			}
			
			distr.set(fDCTMCStateToMDPState[dTMCStatesToFDCTMCState[lastNonZeroIndex]], 1-sum);
			
//			mainLog.println("CREATING DISTRIBUTION: " + distr);
			//TODO allocates more memory and then it releases	
			//add built distribution as a choice to mdp
			if(storeChoices)
				mdp.addChoice(fDCTMCStateToMDPState[fDIndexToEntry[fDIndex]], distr);

			
			//set resultProb as initial distribution for next interval
			temp = resultProb;
			resultProb = transProb;
			transProb = temp;

			if(i%1000000 == 0){ 
				mainLog.println(" " + i + "/" + maxIters );
//				mainLog.println(" " + i + "/" + maxIters + " in time " + ((System.currentTimeMillis()-timer2)/1000.0) + "s.");
//				timer2 = System.currentTimeMillis();
			}
			
		}
		
		timer = System.currentTimeMillis()-timer;
//		mainLog.println("\nTransition probabilities and costs computed in " + (timer/1000.0) + " s.\n");
		
		reward[0] = (fixedDelayReward + weightedExpTransReward + weightedRateReward);
		
		return distr;
		
	} //done computation of this potato
}
